#
# This file is licensed under the Affero General Public License (AGPL) version 3.
#
# Copyright (C) 2023 New Vector, Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# See the GNU Affero General Public License for more details:
# <https://www.gnu.org/licenses/agpl-3.0.html>.
#
# Originally licensed under the Apache License, Version 2.0:
# <http://www.apache.org/licenses/LICENSE-2.0>.
#
# [This file includes modifications made by New Vector Limited]
#
#
from http import HTTPStatus
from typing import Collection, ContextManager, List, Optional
from unittest.mock import AsyncMock, Mock, patch

from parameterized import parameterized, parameterized_class

from twisted.internet import defer
from twisted.test.proto_helpers import MemoryReactor

from synapse.api.constants import AccountDataTypes, EventTypes, JoinRules
from synapse.api.errors import Codes, ResourceLimitError
from synapse.api.filtering import FilterCollection, Filtering
from synapse.api.room_versions import RoomVersion, RoomVersions
from synapse.events import EventBase
from synapse.events.snapshot import EventContext
from synapse.federation.federation_base import event_from_pdu_json
from synapse.handlers.sync import (
    SyncConfig,
    SyncRequestKey,
    SyncResult,
    SyncVersion,
    TimelineBatch,
)
from synapse.rest import admin
from synapse.rest.client import knock, login, room
from synapse.server import HomeServer
from synapse.types import (
    JsonDict,
    MultiWriterStreamToken,
    RoomStreamToken,
    StreamKeyType,
    UserID,
    create_requester,
)
from synapse.util import Clock

import tests.unittest
import tests.utils

_request_key = 0


def generate_request_key() -> SyncRequestKey:
    global _request_key
    _request_key += 1
    return ("request_key", _request_key)


@parameterized_class(
    ("use_state_after",),
    [
        (True,),
        (False,),
    ],
    class_name_func=lambda cls,
    num,
    params_dict: f"{cls.__name__}_{'state_after' if params_dict['use_state_after'] else 'state'}",
)
class SyncTestCase(tests.unittest.HomeserverTestCase):
    """Tests Sync Handler."""

    use_state_after: bool

    servlets = [
        admin.register_servlets,
        knock.register_servlets,
        login.register_servlets,
        room.register_servlets,
    ]

    def prepare(self, reactor: MemoryReactor, clock: Clock, hs: HomeServer) -> None:
        self.sync_handler = self.hs.get_sync_handler()
        self.store = self.hs.get_datastores().main

        # AuthBlocking reads from the hs' config on initialization. We need to
        # modify its config instead of the hs'
        self.auth_blocking = self.hs.get_auth_blocking()

    def test_wait_for_sync_for_user_auth_blocking(self) -> None:
        user_id1 = "@user1:test"
        user_id2 = "@user2:test"
        sync_config = generate_sync_config(
            user_id1, use_state_after=self.use_state_after
        )
        requester = create_requester(user_id1)

        self.reactor.advance(100)  # So we get not 0 time
        self.auth_blocking._limit_usage_by_mau = True
        self.auth_blocking._max_mau_value = 1

        # Check that the happy case does not throw errors
        self.get_success(self.store.upsert_monthly_active_user(user_id1))
        self.get_success(
            self.sync_handler.wait_for_sync_for_user(
                requester,
                sync_config,
                sync_version=SyncVersion.SYNC_V2,
                request_key=generate_request_key(),
            )
        )

        # Test that global lock works
        self.auth_blocking._hs_disabled = True
        e = self.get_failure(
            self.sync_handler.wait_for_sync_for_user(
                requester,
                sync_config,
                sync_version=SyncVersion.SYNC_V2,
                request_key=generate_request_key(),
            ),
            ResourceLimitError,
        )
        self.assertEqual(e.value.errcode, Codes.RESOURCE_LIMIT_EXCEEDED)

        self.auth_blocking._hs_disabled = False

        sync_config = generate_sync_config(
            user_id2, use_state_after=self.use_state_after
        )
        requester = create_requester(user_id2)

        e = self.get_failure(
            self.sync_handler.wait_for_sync_for_user(
                requester,
                sync_config,
                sync_version=SyncVersion.SYNC_V2,
                request_key=generate_request_key(),
            ),
            ResourceLimitError,
        )
        self.assertEqual(e.value.errcode, Codes.RESOURCE_LIMIT_EXCEEDED)

    def test_unknown_room_version(self) -> None:
        """
        A room with an unknown room version should not break sync (and should be excluded).
        """
        inviter = self.register_user("creator", "pass", admin=True)
        inviter_tok = self.login("@creator:test", "pass")

        user = self.register_user("user", "pass")
        tok = self.login("user", "pass")

        # Do an initial sync on a different device.
        requester = create_requester(user)
        initial_result = self.get_success(
            self.sync_handler.wait_for_sync_for_user(
                requester,
                sync_config=generate_sync_config(
                    user, device_id="dev", use_state_after=self.use_state_after
                ),
                sync_version=SyncVersion.SYNC_V2,
                request_key=generate_request_key(),
            )
        )

        # Create a room as the user.
        joined_room = self.helper.create_room_as(user, tok=tok)

        # Invite the user to the room as someone else.
        invite_room = self.helper.create_room_as(inviter, tok=inviter_tok)
        self.helper.invite(invite_room, targ=user, tok=inviter_tok)

        knock_room = self.helper.create_room_as(
            inviter, room_version=RoomVersions.V7.identifier, tok=inviter_tok
        )
        self.helper.send_state(
            knock_room,
            EventTypes.JoinRules,
            {"join_rule": JoinRules.KNOCK},
            tok=inviter_tok,
        )
        channel = self.make_request(
            "POST",
            "/_matrix/client/r0/knock/%s" % (knock_room,),
            b"{}",
            tok,
        )
        self.assertEqual(200, channel.code, channel.result)

        # The rooms should appear in the sync response.
        result = self.get_success(
            self.sync_handler.wait_for_sync_for_user(
                requester,
                sync_config=generate_sync_config(
                    user, use_state_after=self.use_state_after
                ),
                sync_version=SyncVersion.SYNC_V2,
                request_key=generate_request_key(),
            )
        )
        self.assertIn(joined_room, [r.room_id for r in result.joined])
        self.assertIn(invite_room, [r.room_id for r in result.invited])
        self.assertIn(knock_room, [r.room_id for r in result.knocked])

        # Test a incremental sync (by providing a since_token).
        result = self.get_success(
            self.sync_handler.wait_for_sync_for_user(
                requester,
                sync_config=generate_sync_config(
                    user, device_id="dev", use_state_after=self.use_state_after
                ),
                sync_version=SyncVersion.SYNC_V2,
                request_key=generate_request_key(),
                since_token=initial_result.next_batch,
            )
        )
        self.assertIn(joined_room, [r.room_id for r in result.joined])
        self.assertIn(invite_room, [r.room_id for r in result.invited])
        self.assertIn(knock_room, [r.room_id for r in result.knocked])

        # Poke the database and update the room version to an unknown one.
        for room_id in (joined_room, invite_room, knock_room):
            self.get_success(
                self.hs.get_datastores().main.db_pool.simple_update(
                    "rooms",
                    keyvalues={"room_id": room_id},
                    updatevalues={"room_version": "unknown-room-version"},
                    desc="updated-room-version",
                )
            )

        # Blow away caches (supported room versions can only change due to a restart).
        self.store.get_rooms_for_user.invalidate_all()
        self.store._get_rooms_for_local_user_where_membership_is_inner.invalidate_all()
        self.store._get_event_cache.clear()
        self.store._event_ref.clear()

        # The rooms should be excluded from the sync response.
        # Get a new request key.
        result = self.get_success(
            self.sync_handler.wait_for_sync_for_user(
                requester,
                sync_config=generate_sync_config(
                    user, use_state_after=self.use_state_after
                ),
                sync_version=SyncVersion.SYNC_V2,
                request_key=generate_request_key(),
            )
        )
        self.assertNotIn(joined_room, [r.room_id for r in result.joined])
        self.assertNotIn(invite_room, [r.room_id for r in result.invited])
        self.assertNotIn(knock_room, [r.room_id for r in result.knocked])

        # The rooms should also not be in an incremental sync.
        result = self.get_success(
            self.sync_handler.wait_for_sync_for_user(
                requester,
                sync_config=generate_sync_config(
                    user, device_id="dev", use_state_after=self.use_state_after
                ),
                sync_version=SyncVersion.SYNC_V2,
                request_key=generate_request_key(),
                since_token=initial_result.next_batch,
            )
        )
        self.assertNotIn(joined_room, [r.room_id for r in result.joined])
        self.assertNotIn(invite_room, [r.room_id for r in result.invited])
        self.assertNotIn(knock_room, [r.room_id for r in result.knocked])

    def test_ban_wins_race_with_join(self) -> None:
        """Rooms shouldn't appear under "joined" if a join loses a race to a ban.

        A complicated edge case. Imagine the following scenario:

        * you attempt to join a room
        * racing with that is a ban which comes in over federation, which ends up with
          an earlier stream_ordering than the join.
        * you get a sync response with a sync token which is _after_ the ban, but before
          the join
        * now your join lands; it is a valid event because its `prev_event`s predate the
          ban, but will not make it into current_state_events (because bans win over
          joins in state res, essentially).
        * When we do a sync from the incremental sync, the only event in the timeline
          is your join ... and yet you aren't joined.

        The ban coming in over federation isn't crucial for this behaviour; the key
        requirements are:
        1. the homeserver generates a join event with prev_events that precede the ban
           (so that it passes the "are you banned" test)
        2. the join event has a stream_ordering after that of the ban.

        We use monkeypatching to artificially trigger condition (1).
        """
        # A local user Alice creates a room.
        owner = self.register_user("alice", "password")
        owner_tok = self.login(owner, "password")
        room_id = self.helper.create_room_as(owner, is_public=True, tok=owner_tok)

        # Do a sync as Alice to get the latest event in the room.
        alice_sync_result: SyncResult = self.get_success(
            self.sync_handler.wait_for_sync_for_user(
                create_requester(owner),
                generate_sync_config(owner, use_state_after=self.use_state_after),
                sync_version=SyncVersion.SYNC_V2,
                request_key=generate_request_key(),
            )
        )
        self.assertEqual(len(alice_sync_result.joined), 1)
        self.assertEqual(alice_sync_result.joined[0].room_id, room_id)
        last_room_creation_event_id = (
            alice_sync_result.joined[0].timeline.events[-1].event_id
        )

        # Eve, a ne'er-do-well, registers.
        eve = self.register_user("eve", "password")
        eve_token = self.login(eve, "password")

        # Alice preemptively bans Eve.
        self.helper.ban(room_id, owner, eve, tok=owner_tok)

        # Eve syncs.
        eve_requester = create_requester(eve)
        eve_sync_config = generate_sync_config(
            eve, use_state_after=self.use_state_after
        )
        eve_sync_after_ban: SyncResult = self.get_success(
            self.sync_handler.wait_for_sync_for_user(
                eve_requester,
                eve_sync_config,
                sync_version=SyncVersion.SYNC_V2,
                request_key=generate_request_key(),
            )
        )

        # Sanity check this sync result. We shouldn't be joined to the room.
        self.assertEqual(eve_sync_after_ban.joined, [])

        # Eve tries to join the room. We monkey patch the internal logic which selects
        # the prev_events used when creating the join event, such that the ban does not
        # precede the join.
        with self._patch_get_latest_events([last_room_creation_event_id]):
            self.helper.join(
                room_id,
                eve,
                tok=eve_token,
                # Previously, this join would succeed but now we expect it to fail at
                # this point. The rest of the test is for the case when this used to
                # succeed.
                expect_code=HTTPStatus.FORBIDDEN,
            )

        # Eve makes a second, incremental sync.
        eve_incremental_sync_after_join: SyncResult = self.get_success(
            self.sync_handler.wait_for_sync_for_user(
                eve_requester,
                eve_sync_config,
                sync_version=SyncVersion.SYNC_V2,
                request_key=generate_request_key(),
                since_token=eve_sync_after_ban.next_batch,
            )
        )
        # Eve should not see herself as joined to the room.
        self.assertEqual(eve_incremental_sync_after_join.joined, [])

        # If we did a third initial sync, we should _still_ see eve is not joined to the room.
        eve_initial_sync_after_join: SyncResult = self.get_success(
            self.sync_handler.wait_for_sync_for_user(
                eve_requester,
                eve_sync_config,
                sync_version=SyncVersion.SYNC_V2,
                request_key=generate_request_key(),
                since_token=None,
            )
        )
        self.assertEqual(eve_initial_sync_after_join.joined, [])

    def test_state_includes_changes_on_forks(self) -> None:
        """State changes that happen on a fork of the DAG must be included in `state`

        Given the following DAG:

             E1
           ↗    ↖
          |      S2
          |      ↑
        --|------|----
          |      |
          E3     |
           ↖    /
             E4

        ... and a filter that means we only return 2 events, represented by the dashed
        horizontal line: `S2` must be included in the `state` section.
        """
        alice = self.register_user("alice", "password")
        alice_tok = self.login(alice, "password")
        alice_requester = create_requester(alice)
        room_id = self.helper.create_room_as(alice, is_public=True, tok=alice_tok)

        # Do an initial sync as Alice to get a known starting point.
        initial_sync_result = self.get_success(
            self.sync_handler.wait_for_sync_for_user(
                alice_requester,
                generate_sync_config(alice, use_state_after=self.use_state_after),
                sync_version=SyncVersion.SYNC_V2,
                request_key=generate_request_key(),
            )
        )
        last_room_creation_event_id = (
            initial_sync_result.joined[0].timeline.events[-1].event_id
        )

        # Send a state event, and a regular event, both using the same prev ID
        with self._patch_get_latest_events([last_room_creation_event_id]):
            s2_event = self.helper.send_state(room_id, "s2", {}, tok=alice_tok)[
                "event_id"
            ]
            e3_event = self.helper.send(room_id, "e3", tok=alice_tok)["event_id"]

        # Send a final event, joining the two branches of the dag
        e4_event = self.helper.send(room_id, "e4", tok=alice_tok)["event_id"]

        # do an incremental sync, with a filter that will ensure we only get two of
        # the three new events.
        incremental_sync = self.get_success(
            self.sync_handler.wait_for_sync_for_user(
                alice_requester,
                generate_sync_config(
                    alice,
                    filter_collection=FilterCollection(
                        self.hs, {"room": {"timeline": {"limit": 2}}}
                    ),
                    use_state_after=self.use_state_after,
                ),
                sync_version=SyncVersion.SYNC_V2,
                request_key=generate_request_key(),
                since_token=initial_sync_result.next_batch,
            )
        )

        # The state event should appear in the 'state' section of the response.
        room_sync = incremental_sync.joined[0]
        self.assertEqual(room_sync.room_id, room_id)
        self.assertTrue(room_sync.timeline.limited)
        self.assertEqual(
            [e.event_id for e in room_sync.timeline.events],
            [e3_event, e4_event],
        )
        self.assertEqual(
            [e.event_id for e in room_sync.state.values()],
            [s2_event],
        )

    def test_state_includes_changes_on_forks_when_events_excluded(self) -> None:
        """A variation on the previous test, but where one event is filtered

        The DAG is the same as the previous test, but E4 is excluded by the filter.

             E1
           ↗    ↖
          |      S2
          |      ↑
        --|------|----
          |      |
          E3     |
           ↖    /
            (E4)

        """

        alice = self.register_user("alice", "password")
        alice_tok = self.login(alice, "password")
        alice_requester = create_requester(alice)
        room_id = self.helper.create_room_as(alice, is_public=True, tok=alice_tok)

        # Do an initial sync as Alice to get a known starting point.
        initial_sync_result = self.get_success(
            self.sync_handler.wait_for_sync_for_user(
                alice_requester,
                generate_sync_config(alice, use_state_after=self.use_state_after),
                sync_version=SyncVersion.SYNC_V2,
                request_key=generate_request_key(),
            )
        )
        last_room_creation_event_id = (
            initial_sync_result.joined[0].timeline.events[-1].event_id
        )

        # Send a state event, and a regular event, both using the same prev ID
        with self._patch_get_latest_events([last_room_creation_event_id]):
            s2_event = self.helper.send_state(room_id, "s2", {}, tok=alice_tok)[
                "event_id"
            ]
            e3_event = self.helper.send(room_id, "e3", tok=alice_tok)["event_id"]

        # Send a final event, joining the two branches of the dag
        self.helper.send(room_id, "e4", type="not_a_normal_message", tok=alice_tok)[
            "event_id"
        ]

        # do an incremental sync, with a filter that will only return E3, excluding S2
        # and E4.
        incremental_sync = self.get_success(
            self.sync_handler.wait_for_sync_for_user(
                alice_requester,
                generate_sync_config(
                    alice,
                    filter_collection=FilterCollection(
                        self.hs,
                        {
                            "room": {
                                "timeline": {
                                    "limit": 1,
                                    "not_types": ["not_a_normal_message"],
                                }
                            }
                        },
                    ),
                    use_state_after=self.use_state_after,
                ),
                sync_version=SyncVersion.SYNC_V2,
                request_key=generate_request_key(),
                since_token=initial_sync_result.next_batch,
            )
        )

        # The state event should appear in the 'state' section of the response.
        room_sync = incremental_sync.joined[0]
        self.assertEqual(room_sync.room_id, room_id)
        self.assertTrue(room_sync.timeline.limited)
        self.assertEqual(
            [e.event_id for e in room_sync.timeline.events],
            [e3_event],
        )
        self.assertEqual(
            [e.event_id for e in room_sync.state.values()],
            [s2_event],
        )

    def test_state_includes_changes_on_long_lived_forks(self) -> None:
        """State changes that happen on a fork of the DAG must be included in `state`

        Given the following DAG:

             E1
           ↗    ↖
          |      S2
          |      ↑
        --|------|----
          E3     |
        --|------|----
          |      E4
          |      |

        ... and a filter that means we only return 1 event, represented by the dashed
        horizontal lines: `S2` must be included in the `state` section on the second sync.

        When `use_state_after` is enabled, then we expect to see `s2` in the first sync.
        """
        alice = self.register_user("alice", "password")
        alice_tok = self.login(alice, "password")
        alice_requester = create_requester(alice)
        room_id = self.helper.create_room_as(alice, is_public=True, tok=alice_tok)

        # Do an initial sync as Alice to get a known starting point.
        initial_sync_result = self.get_success(
            self.sync_handler.wait_for_sync_for_user(
                alice_requester,
                generate_sync_config(alice, use_state_after=self.use_state_after),
                sync_version=SyncVersion.SYNC_V2,
                request_key=generate_request_key(),
            )
        )
        last_room_creation_event_id = (
            initial_sync_result.joined[0].timeline.events[-1].event_id
        )

        # Send a state event, and a regular event, both using the same prev ID
        with self._patch_get_latest_events([last_room_creation_event_id]):
            s2_event = self.helper.send_state(room_id, "s2", {}, tok=alice_tok)[
                "event_id"
            ]
            e3_event = self.helper.send(room_id, "e3", tok=alice_tok)["event_id"]

        # Do an incremental sync, this will return E3 but *not* S2 at this
        # point.
        incremental_sync = self.get_success(
            self.sync_handler.wait_for_sync_for_user(
                alice_requester,
                generate_sync_config(
                    alice,
                    filter_collection=FilterCollection(
                        self.hs, {"room": {"timeline": {"limit": 1}}}
                    ),
                    use_state_after=self.use_state_after,
                ),
                sync_version=SyncVersion.SYNC_V2,
                request_key=generate_request_key(),
                since_token=initial_sync_result.next_batch,
            )
        )
        room_sync = incremental_sync.joined[0]
        self.assertEqual(room_sync.room_id, room_id)
        self.assertTrue(room_sync.timeline.limited)
        self.assertEqual(
            [e.event_id for e in room_sync.timeline.events],
            [e3_event],
        )

        if self.use_state_after:
            # When using `state_after` we get told about s2 immediately
            self.assertEqual(
                [e.event_id for e in room_sync.state.values()],
                [s2_event],
            )
        else:
            self.assertEqual(
                [e.event_id for e in room_sync.state.values()],
                [],
            )

        # Now send another event that points to S2, but not E3.
        with self._patch_get_latest_events([s2_event]):
            e4_event = self.helper.send(room_id, "e4", tok=alice_tok)["event_id"]

        # Doing an incremental sync should return S2 in state.
        incremental_sync = self.get_success(
            self.sync_handler.wait_for_sync_for_user(
                alice_requester,
                generate_sync_config(
                    alice,
                    filter_collection=FilterCollection(
                        self.hs, {"room": {"timeline": {"limit": 1}}}
                    ),
                    use_state_after=self.use_state_after,
                ),
                sync_version=SyncVersion.SYNC_V2,
                request_key=generate_request_key(),
                since_token=incremental_sync.next_batch,
            )
        )
        room_sync = incremental_sync.joined[0]
        self.assertEqual(room_sync.room_id, room_id)
        self.assertFalse(room_sync.timeline.limited)
        self.assertEqual(
            [e.event_id for e in room_sync.timeline.events],
            [e4_event],
        )

        if self.use_state_after:
            # When using `state_after` we got told about s2 previously, so we
            # don't again.
            self.assertEqual(
                [e.event_id for e in room_sync.state.values()],
                [],
            )
        else:
            self.assertEqual(
                [e.event_id for e in room_sync.state.values()],
                [s2_event],
            )

    def test_state_includes_changes_on_ungappy_syncs(self) -> None:
        """Test `state` where the sync is not gappy.

        We start with a DAG like this:

             E1
           ↗    ↖
          |      S2
          |
        --|---
          |
          E3

        ... and initialsync with `limit=1`, represented by the horizontal dashed line.
        At this point, we do not expect S2 to appear in the response at all (since
        it is excluded from the timeline by the `limit`, and the state is based on the
        state after the most recent event before the sync token (E3), which doesn't
        include S2.

        Now more events arrive, and we do an incremental sync:

             E1
           ↗    ↖
          |      S2
          |      ↑
          E3     |
          ↑      |
        --|------|----
          |      |
          E4     |
           ↖    /
             E5

        This is the last chance for us to tell the client about S2, so it *must* be
        included in the response.

        When `use_state_after` is enabled, then we expect to see `s2` in the first sync.
        """
        alice = self.register_user("alice", "password")
        alice_tok = self.login(alice, "password")
        alice_requester = create_requester(alice)
        room_id = self.helper.create_room_as(alice, is_public=True, tok=alice_tok)

        # Do an initial sync to get a known starting point.
        initial_sync_result = self.get_success(
            self.sync_handler.wait_for_sync_for_user(
                alice_requester,
                generate_sync_config(alice, use_state_after=self.use_state_after),
                sync_version=SyncVersion.SYNC_V2,
                request_key=generate_request_key(),
            )
        )
        last_room_creation_event_id = (
            initial_sync_result.joined[0].timeline.events[-1].event_id
        )

        # Send a state event, and a regular event, both using the same prev ID
        with self._patch_get_latest_events([last_room_creation_event_id]):
            s2_event = self.helper.send_state(room_id, "s2", {}, tok=alice_tok)[
                "event_id"
            ]
            e3_event = self.helper.send(room_id, "e3", tok=alice_tok)["event_id"]

        # Another initial sync, with limit=1
        initial_sync_result = self.get_success(
            self.sync_handler.wait_for_sync_for_user(
                alice_requester,
                generate_sync_config(
                    alice,
                    filter_collection=FilterCollection(
                        self.hs, {"room": {"timeline": {"limit": 1}}}
                    ),
                    use_state_after=self.use_state_after,
                ),
                sync_version=SyncVersion.SYNC_V2,
                request_key=generate_request_key(),
            )
        )
        room_sync = initial_sync_result.joined[0]
        self.assertEqual(room_sync.room_id, room_id)
        self.assertEqual(
            [e.event_id for e in room_sync.timeline.events],
            [e3_event],
        )
        if self.use_state_after:
            # When using `state_after` we get told about s2 immediately
            self.assertIn(s2_event, [e.event_id for e in room_sync.state.values()])
        else:
            self.assertNotIn(s2_event, [e.event_id for e in room_sync.state.values()])

        # More events, E4 and E5
        with self._patch_get_latest_events([e3_event]):
            e4_event = self.helper.send(room_id, "e4", tok=alice_tok)["event_id"]
        e5_event = self.helper.send(room_id, "e5", tok=alice_tok)["event_id"]

        # Now incremental sync
        incremental_sync = self.get_success(
            self.sync_handler.wait_for_sync_for_user(
                alice_requester,
                generate_sync_config(alice, use_state_after=self.use_state_after),
                sync_version=SyncVersion.SYNC_V2,
                request_key=generate_request_key(),
                since_token=initial_sync_result.next_batch,
            )
        )

        # The state event should appear in the 'state' section of the response.
        room_sync = incremental_sync.joined[0]
        self.assertEqual(room_sync.room_id, room_id)
        self.assertFalse(room_sync.timeline.limited)
        self.assertEqual(
            [e.event_id for e in room_sync.timeline.events],
            [e4_event, e5_event],
        )

        if self.use_state_after:
            # When using `state_after` we got told about s2 previously, so we
            # don't again.
            self.assertEqual(
                [e.event_id for e in room_sync.state.values()],
                [],
            )
        else:
            self.assertEqual(
                [e.event_id for e in room_sync.state.values()],
                [s2_event],
            )

    @parameterized.expand(
        [
            (False, False),
            (True, False),
            (False, True),
            (True, True),
        ],
        name_func=lambda func, num, p: f"{func.__name__}_{p.args[0]}_{p.args[1]}",
    )
    def test_archived_rooms_do_not_include_state_after_leave(
        self, initial_sync: bool, empty_timeline: bool
    ) -> None:
        """If the user leaves the room, state changes that happen after they leave are not returned.

        We try with both a zero and a normal timeline limit,
        and we try both an initial sync and an incremental sync for both.
        """
        if empty_timeline and not initial_sync:
            # FIXME synapse doesn't return the room at all in this situation!
            self.skipTest("Synapse does not correctly handle this case")

        # Alice creates the room, and bob joins.
        alice = self.register_user("alice", "password")
        alice_tok = self.login(alice, "password")

        bob = self.register_user("bob", "password")
        bob_tok = self.login(bob, "password")
        bob_requester = create_requester(bob)

        room_id = self.helper.create_room_as(alice, is_public=True, tok=alice_tok)
        self.helper.join(room_id, bob, tok=bob_tok)

        initial_sync_result = self.get_success(
            self.sync_handler.wait_for_sync_for_user(
                bob_requester,
                generate_sync_config(bob, use_state_after=self.use_state_after),
                sync_version=SyncVersion.SYNC_V2,
                request_key=generate_request_key(),
            )
        )

        # Alice sends a message and a state
        before_message_event = self.helper.send(room_id, "before", tok=alice_tok)[
            "event_id"
        ]
        before_state_event = self.helper.send_state(
            room_id, "test_state", {"body": "before"}, tok=alice_tok
        )["event_id"]

        # Bob leaves
        leave_event = self.helper.leave(room_id, bob, tok=bob_tok)["event_id"]

        # Alice sends some more stuff
        self.helper.send(room_id, "after", tok=alice_tok)["event_id"]
        self.helper.send_state(room_id, "test_state", {"body": "after"}, tok=alice_tok)[
            "event_id"
        ]

        # And now, Bob resyncs.
        filter_dict: JsonDict = {"room": {"include_leave": True}}
        if empty_timeline:
            filter_dict["room"]["timeline"] = {"limit": 0}
        sync_room_result = self.get_success(
            self.sync_handler.wait_for_sync_for_user(
                bob_requester,
                generate_sync_config(
                    bob,
                    filter_collection=FilterCollection(self.hs, filter_dict),
                    use_state_after=self.use_state_after,
                ),
                sync_version=SyncVersion.SYNC_V2,
                request_key=generate_request_key(),
                since_token=None if initial_sync else initial_sync_result.next_batch,
            )
        ).archived[0]

        if empty_timeline:
            # The timeline should be empty
            self.assertEqual(sync_room_result.timeline.events, [])
        else:
            # The last three events in the timeline should be those leading up to the
            # leave
            self.assertEqual(
                [e.event_id for e in sync_room_result.timeline.events[-3:]],
                [before_message_event, before_state_event, leave_event],
            )

        if empty_timeline or self.use_state_after:
            # And the state should include the leave event...
            self.assertEqual(
                sync_room_result.state[("m.room.member", bob)].event_id, leave_event
            )
            # ... and the state change before he left.
            self.assertEqual(
                sync_room_result.state[("test_state", "")].event_id, before_state_event
            )
        else:
            # ... And the state should be empty
            self.assertEqual(sync_room_result.state, {})

    def _patch_get_latest_events(self, latest_events: List[str]) -> ContextManager:
        """Monkey-patch `get_prev_events_for_room`

        Returns a context manager which will replace the implementation of
        `get_prev_events_for_room` with one which returns `latest_events`.
        """
        return patch.object(
            self.hs.get_datastores().main,
            "get_prev_events_for_room",
            new_callable=AsyncMock,
            return_value=latest_events,
        )

    def test_call_invite_in_public_room_not_returned(self) -> None:
        user = self.register_user("alice", "password")
        tok = self.login(user, "password")
        room_id = self.helper.create_room_as(user, is_public=True, tok=tok)
        self.handler = self.hs.get_federation_handler()
        federation_event_handler = self.hs.get_federation_event_handler()

        async def _check_event_auth(
            origin: Optional[str], event: EventBase, context: EventContext
        ) -> None:
            pass

        federation_event_handler._check_event_auth = _check_event_auth  # type: ignore[method-assign]
        self.client = self.hs.get_federation_client()

        async def _check_sigs_and_hash_for_pulled_events_and_fetch(
            dest: str, pdus: Collection[EventBase], room_version: RoomVersion
        ) -> List[EventBase]:
            return list(pdus)

        self.client._check_sigs_and_hash_for_pulled_events_and_fetch = (  # type: ignore[method-assign]
            _check_sigs_and_hash_for_pulled_events_and_fetch  # type: ignore[assignment]
        )

        prev_events = self.get_success(self.store.get_prev_events_for_room(room_id))

        # create a call invite event
        call_event = event_from_pdu_json(
            {
                "type": EventTypes.CallInvite,
                "content": {},
                "room_id": room_id,
                "sender": user,
                "depth": 32,
                "prev_events": prev_events,
                "auth_events": prev_events,
                "origin_server_ts": self.clock.time_msec(),
            },
            RoomVersions.V10,
        )

        self.assertEqual(
            self.get_success(
                federation_event_handler.on_receive_pdu("test.serv", call_event)
            ),
            None,
        )

        # check that it is in DB
        recent_event = self.get_success(self.store.get_prev_events_for_room(room_id))
        self.assertIn(call_event.event_id, recent_event)

        # but that it does not come down /sync in public room
        sync_result: SyncResult = self.get_success(
            self.sync_handler.wait_for_sync_for_user(
                create_requester(user),
                generate_sync_config(user, use_state_after=self.use_state_after),
                sync_version=SyncVersion.SYNC_V2,
                request_key=generate_request_key(),
            )
        )
        event_ids = []
        for event in sync_result.joined[0].timeline.events:
            event_ids.append(event.event_id)
        self.assertNotIn(call_event.event_id, event_ids)

        # it will come down in a private room, though
        user2 = self.register_user("bob", "password")
        tok2 = self.login(user2, "password")
        private_room_id = self.helper.create_room_as(
            user2, is_public=False, tok=tok2, extra_content={"preset": "private_chat"}
        )

        priv_prev_events = self.get_success(
            self.store.get_prev_events_for_room(private_room_id)
        )
        private_call_event = event_from_pdu_json(
            {
                "type": EventTypes.CallInvite,
                "content": {},
                "room_id": private_room_id,
                "sender": user,
                "depth": 32,
                "prev_events": priv_prev_events,
                "auth_events": priv_prev_events,
                "origin_server_ts": self.clock.time_msec(),
            },
            RoomVersions.V10,
        )

        self.assertEqual(
            self.get_success(
                federation_event_handler.on_receive_pdu("test.serv", private_call_event)
            ),
            None,
        )

        recent_events = self.get_success(
            self.store.get_prev_events_for_room(private_room_id)
        )
        self.assertIn(private_call_event.event_id, recent_events)

        private_sync_result: SyncResult = self.get_success(
            self.sync_handler.wait_for_sync_for_user(
                create_requester(user2),
                generate_sync_config(user2, use_state_after=self.use_state_after),
                sync_version=SyncVersion.SYNC_V2,
                request_key=generate_request_key(),
            )
        )
        priv_event_ids = []
        for event in private_sync_result.joined[0].timeline.events:
            priv_event_ids.append(event.event_id)

        self.assertIn(private_call_event.event_id, priv_event_ids)

    def test_push_rules_with_bad_account_data(self) -> None:
        """Some old accounts have managed to set a `m.push_rules` account data,
        which we should ignore in /sync response.
        """

        user = self.register_user("alice", "password")

        # Insert the bad account data.
        self.get_success(
            self.store.add_account_data_for_user(user, AccountDataTypes.PUSH_RULES, {})
        )

        sync_result: SyncResult = self.get_success(
            self.sync_handler.wait_for_sync_for_user(
                create_requester(user),
                generate_sync_config(user, use_state_after=self.use_state_after),
                sync_version=SyncVersion.SYNC_V2,
                request_key=generate_request_key(),
            )
        )

        for account_dict in sync_result.account_data:
            if account_dict["type"] == AccountDataTypes.PUSH_RULES:
                # We should have lots of push rules here, rather than the bad
                # empty data.
                self.assertNotEqual(account_dict["content"], {})
                return

        self.fail("No push rules found")

    def test_wait_for_future_sync_token(self) -> None:
        """Test that if we receive a token that is ahead of our current token,
        we'll wait until the stream position advances.

        This can happen if replication streams start lagging, and the client's
        previous sync request was serviced by a worker ahead of ours.
        """
        user = self.register_user("alice", "password")

        # We simulate a lagging stream by getting a stream ID from the ID gen
        # and then waiting to mark it as "persisted".
        presence_id_gen = self.store.get_presence_stream_id_gen()
        ctx_mgr = presence_id_gen.get_next()
        stream_id = self.get_success(ctx_mgr.__aenter__())

        # Create the new token based on the stream ID above.
        current_token = self.hs.get_event_sources().get_current_token()
        since_token = current_token.copy_and_advance(StreamKeyType.PRESENCE, stream_id)

        sync_d = defer.ensureDeferred(
            self.sync_handler.wait_for_sync_for_user(
                create_requester(user),
                generate_sync_config(user, use_state_after=self.use_state_after),
                sync_version=SyncVersion.SYNC_V2,
                request_key=generate_request_key(),
                since_token=since_token,
                timeout=0,
            )
        )

        # This should block waiting for the presence stream to update
        self.pump()
        self.assertFalse(sync_d.called)

        # Marking the stream ID as persisted should unblock the request.
        self.get_success(ctx_mgr.__aexit__(None, None, None))

        self.get_success(sync_d, by=1.0)

    @parameterized.expand(
        [(key,) for key in StreamKeyType.__members__.values()],
        name_func=lambda func, _, param: f"{func.__name__}_{param.args[0].name}",
    )
    def test_wait_for_invalid_future_sync_token(
        self, stream_key: StreamKeyType
    ) -> None:
        """Like the previous test, except we give a token that has a stream
        position ahead of what is in the DB, i.e. its invalid and we shouldn't
        wait for the stream to advance (as it may never do so).

        This can happen due to older versions of Synapse giving out stream
        positions without persisting them in the DB, and so on restart the
        stream would get reset back to an older position.
        """
        user = self.register_user("alice", "password")

        # Create a token and advance one of the streams.
        current_token = self.hs.get_event_sources().get_current_token()
        token_value = current_token.get_field(stream_key)

        # How we advance the streams depends on the type.
        if isinstance(token_value, int):
            since_token = current_token.copy_and_advance(stream_key, token_value + 1)
        elif isinstance(token_value, MultiWriterStreamToken):
            since_token = current_token.copy_and_advance(
                stream_key, MultiWriterStreamToken(stream=token_value.stream + 1)
            )
        elif isinstance(token_value, RoomStreamToken):
            since_token = current_token.copy_and_advance(
                stream_key, RoomStreamToken(stream=token_value.stream + 1)
            )
        else:
            raise Exception("Unreachable")

        sync_d = defer.ensureDeferred(
            self.sync_handler.wait_for_sync_for_user(
                create_requester(user),
                generate_sync_config(user, use_state_after=self.use_state_after),
                sync_version=SyncVersion.SYNC_V2,
                request_key=generate_request_key(),
                since_token=since_token,
                timeout=0,
            )
        )

        # We should return without waiting for the presence stream to advance.
        self.get_success(sync_d)


def generate_sync_config(
    user_id: str,
    device_id: Optional[str] = "device_id",
    filter_collection: Optional[FilterCollection] = None,
    use_state_after: bool = False,
) -> SyncConfig:
    """Generate a sync config (with a unique request key).

    Args:
        user_id: user who is syncing.
        device_id: device that is syncing. Defaults to "device_id".
        filter_collection: filter to apply. Defaults to the default filter (ie,
            return everything, with a default limit)
        use_state_after: whether the `use_state_after` flag was set.
    """
    if filter_collection is None:
        filter_collection = Filtering(Mock()).DEFAULT_FILTER_COLLECTION

    return SyncConfig(
        user=UserID.from_string(user_id),
        filter_collection=filter_collection,
        is_guest=False,
        device_id=device_id,
        use_state_after=use_state_after,
    )


class SyncStateAfterTestCase(tests.unittest.HomeserverTestCase):
    """Tests Sync Handler state behavior when using `use_state_after."""

    servlets = [
        admin.register_servlets,
        knock.register_servlets,
        login.register_servlets,
        room.register_servlets,
    ]

    def prepare(self, reactor: MemoryReactor, clock: Clock, hs: HomeServer) -> None:
        self.sync_handler = self.hs.get_sync_handler()
        self.store = self.hs.get_datastores().main

        # AuthBlocking reads from the hs' config on initialization. We need to
        # modify its config instead of the hs'
        self.auth_blocking = self.hs.get_auth_blocking()

    def test_initial_sync_multiple_deltas(self) -> None:
        """Test that if multiple state deltas have happened during processing of
        a full state sync we return the correct state"""

        user = self.register_user("user", "password")
        tok = self.login("user", "password")

        # Create a room as the user and set some custom state.
        joined_room = self.helper.create_room_as(user, tok=tok)

        first_state = self.helper.send_state(
            joined_room, event_type="m.test_event", body={"num": 1}, tok=tok
        )

        # Take a snapshot of the stream token, to simulate doing an initial sync
        # at this point.
        end_stream_token = self.hs.get_event_sources().get_current_token()

        # Send some state *after* the stream token
        self.helper.send_state(
            joined_room, event_type="m.test_event", body={"num": 2}, tok=tok
        )

        # Calculating the full state will return the first state, and not the
        # second.
        state = self.get_success(
            self.sync_handler._compute_state_delta_for_full_sync(
                room_id=joined_room,
                sync_config=generate_sync_config(user, use_state_after=True),
                batch=TimelineBatch(
                    prev_batch=end_stream_token, events=[], limited=True
                ),
                end_token=end_stream_token,
                members_to_fetch=None,
                timeline_state={},
                joined=True,
            )
        )
        self.assertEqual(state[("m.test_event", "")], first_state["event_id"])

    def test_incremental_sync_multiple_deltas(self) -> None:
        """Test that if multiple state deltas have happened since an incremental
        state sync we return the correct state"""

        user = self.register_user("user", "password")
        tok = self.login("user", "password")

        # Create a room as the user and set some custom state.
        joined_room = self.helper.create_room_as(user, tok=tok)

        # Take a snapshot of the stream token, to simulate doing an incremental sync
        # from this point.
        since_token = self.hs.get_event_sources().get_current_token()

        self.helper.send_state(
            joined_room, event_type="m.test_event", body={"num": 1}, tok=tok
        )

        # Send some state *after* the stream token
        second_state = self.helper.send_state(
            joined_room, event_type="m.test_event", body={"num": 2}, tok=tok
        )

        end_stream_token = self.hs.get_event_sources().get_current_token()

        # Calculating the incrementals state will return the second state, and not the
        # first.
        state = self.get_success(
            self.sync_handler._compute_state_delta_for_incremental_sync(
                room_id=joined_room,
                sync_config=generate_sync_config(user, use_state_after=True),
                batch=TimelineBatch(
                    prev_batch=end_stream_token, events=[], limited=True
                ),
                since_token=since_token,
                end_token=end_stream_token,
                members_to_fetch=None,
                timeline_state={},
            )
        )
        self.assertEqual(state[("m.test_event", "")], second_state["event_id"])

    def test_incremental_sync_lazy_loaded_no_timeline(self) -> None:
        """Test that lazy-loading with an empty timeline doesn't return the full
        state.

        There was a bug where an empty state filter would cause the DB to return
        the full state, rather than an empty set.
        """
        user = self.register_user("user", "password")
        tok = self.login("user", "password")

        # Create a room as the user and set some custom state.
        joined_room = self.helper.create_room_as(user, tok=tok)

        since_token = self.hs.get_event_sources().get_current_token()
        end_stream_token = self.hs.get_event_sources().get_current_token()

        state = self.get_success(
            self.sync_handler._compute_state_delta_for_incremental_sync(
                room_id=joined_room,
                sync_config=generate_sync_config(user, use_state_after=True),
                batch=TimelineBatch(
                    prev_batch=end_stream_token, events=[], limited=True
                ),
                since_token=since_token,
                end_token=end_stream_token,
                members_to_fetch=set(),
                timeline_state={},
            )
        )

        self.assertEqual(state, {})



#
# This file is licensed under the Affero General Public License (AGPL) version 3.
#
# Copyright (C) 2024 New Vector, Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# See the GNU Affero General Public License for more details:
# <https://www.gnu.org/licenses/agpl-3.0.html>.

from typing import Dict

from twisted.web.resource import Resource

from synapse.rest import admin
from synapse.rest.client import login
from synapse.rest.synapse.client import build_synapse_client_resource_tree

from tests import unittest


class FederationWhitelistTests(unittest.HomeserverTestCase):
    servlets = [
        admin.register_servlets_for_client_rest_resource,
        login.register_servlets,
    ]

    def create_resource_dict(self) -> Dict[str, Resource]:
        base = super().create_resource_dict()
        base.update(build_synapse_client_resource_tree(self.hs))
        return base

    def test_default(self) -> None:
        "If the config option is not enabled, the endpoint should 404"
        channel = self.make_request(
            "GET", "/_synapse/client/v1/config/federation_whitelist", shorthand=False
        )

        self.assertEqual(channel.code, 404)

    @unittest.override_config({"federation_whitelist_endpoint_enabled": True})
    def test_no_auth(self) -> None:
        "Endpoint requires auth when enabled"

        channel = self.make_request(
            "GET", "/_synapse/client/v1/config/federation_whitelist", shorthand=False
        )

        self.assertEqual(channel.code, 401)

    @unittest.override_config({"federation_whitelist_endpoint_enabled": True})
    def test_no_whitelist(self) -> None:
        "Test when there is no whitelist configured"

        self.register_user("user", "password")
        tok = self.login("user", "password")

        channel = self.make_request(
            "GET",
            "/_synapse/client/v1/config/federation_whitelist",
            shorthand=False,
            access_token=tok,
        )

        self.assertEqual(channel.code, 200)
        self.assertEqual(
            channel.json_body, {"whitelist_enabled": False, "whitelist": []}
        )

    @unittest.override_config(
        {
            "federation_whitelist_endpoint_enabled": True,
            "federation_domain_whitelist": ["example.com"],
        }
    )
    def test_whitelist(self) -> None:
        "Test when there is a whitelist configured"

        self.register_user("user", "password")
        tok = self.login("user", "password")

        channel = self.make_request(
            "GET",
            "/_synapse/client/v1/config/federation_whitelist",
            shorthand=False,
            access_token=tok,
        )

        self.assertEqual(channel.code, 200)
        self.assertEqual(
            channel.json_body, {"whitelist_enabled": True, "whitelist": ["example.com"]}
        )

    @unittest.override_config(
        {
            "federation_whitelist_endpoint_enabled": True,
            "federation_domain_whitelist": ["example.com", "example.com"],
        }
    )
    def test_whitelist_no_duplicates(self) -> None:
        "Test when there is a whitelist configured with duplicates, no duplicates are returned"

        self.register_user("user", "password")
        tok = self.login("user", "password")

        channel = self.make_request(
            "GET",
            "/_synapse/client/v1/config/federation_whitelist",
            shorthand=False,
            access_token=tok,
        )

        self.assertEqual(channel.code, 200)
        self.assertEqual(
            channel.json_body, {"whitelist_enabled": True, "whitelist": ["example.com"]}
        )



#
# This file is licensed under the Affero General Public License (AGPL) version 3.
#
# Copyright 2021 The Matrix.org Foundation C.I.C.
# Copyright 2015, 2016 OpenMarket Ltd
# Copyright (C) 2023 New Vector, Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# See the GNU Affero General Public License for more details:
# <https://www.gnu.org/licenses/agpl-3.0.html>.
#
# Originally licensed under the Apache License, Version 2.0:
# <http://www.apache.org/licenses/LICENSE-2.0>.
#
# [This file includes modifications made by New Vector Limited]
#
#

from typing import Any, List, Optional

from synapse.config.sso import SsoAttributeRequirement
from synapse.types import JsonDict

from ._base import Config, ConfigError
from ._util import validate_config


class CasConfig(Config):
    """Cas Configuration

    cas_server_url: URL of CAS server
    """

    section = "cas"

    def read_config(self, config: JsonDict, **kwargs: Any) -> None:
        cas_config = config.get("cas_config", None)
        self.cas_enabled = cas_config and cas_config.get("enabled", True)

        if self.cas_enabled:
            self.cas_server_url = cas_config["server_url"]

            # TODO Update this to a _synapse URL.
            public_baseurl = self.root.server.public_baseurl
            self.cas_service_url: Optional[str] = (
                public_baseurl + "_matrix/client/r0/login/cas/ticket"
            )

            self.cas_protocol_version = cas_config.get("protocol_version")
            if (
                self.cas_protocol_version is not None
                and self.cas_protocol_version not in [1, 2, 3]
            ):
                raise ConfigError(
                    "Unsupported CAS protocol version %s (only versions 1, 2, 3 are supported)"
                    % (self.cas_protocol_version,),
                    ("cas_config", "protocol_version"),
                )
            self.cas_displayname_attribute = cas_config.get("displayname_attribute")
            required_attributes = cas_config.get("required_attributes") or {}
            self.cas_required_attributes = _parsed_required_attributes_def(
                required_attributes
            )

            self.cas_enable_registration = cas_config.get("enable_registration", True)

            self.cas_allow_numeric_ids = cas_config.get("allow_numeric_ids")
            self.cas_numeric_ids_prefix = cas_config.get("numeric_ids_prefix")
            if (
                self.cas_numeric_ids_prefix is not None
                and self.cas_numeric_ids_prefix.isalnum() is False
            ):
                raise ConfigError(
                    "Only alphanumeric characters are allowed for numeric IDs prefix",
                    ("cas_config", "numeric_ids_prefix"),
                )

            self.idp_name = cas_config.get("idp_name", "CAS")
            self.idp_icon = cas_config.get("idp_icon")
            self.idp_brand = cas_config.get("idp_brand")

        else:
            self.cas_server_url = None
            self.cas_service_url = None
            self.cas_protocol_version = None
            self.cas_displayname_attribute = None
            self.cas_required_attributes = []
            self.cas_enable_registration = False
            self.cas_allow_numeric_ids = False
            self.cas_numeric_ids_prefix = "u"


# CAS uses a legacy required attributes mapping, not the one provided by
# SsoAttributeRequirement.
REQUIRED_ATTRIBUTES_SCHEMA = {
    "type": "object",
    "additionalProperties": {"anyOf": [{"type": "string"}, {"type": "null"}]},
}


def _parsed_required_attributes_def(
    required_attributes: Any,
) -> List[SsoAttributeRequirement]:
    validate_config(
        REQUIRED_ATTRIBUTES_SCHEMA,
        required_attributes,
        config_path=("cas_config", "required_attributes"),
    )
    return [SsoAttributeRequirement(k, v) for k, v in required_attributes.items()]



#
# This file is licensed under the Affero General Public License (AGPL) version 3.
#
#  Copyright 2021 The Matrix.org Foundation C.I.C.
# Copyright (C) 2023 New Vector, Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# See the GNU Affero General Public License for more details:
# <https://www.gnu.org/licenses/agpl-3.0.html>.
#
# Originally licensed under the Apache License, Version 2.0:
# <http://www.apache.org/licenses/LICENSE-2.0>.
#
# [This file includes modifications made by New Vector Limited]
#
#

import json
from typing import Any

from parameterized import parameterized

from twisted.test.proto_helpers import MemoryReactor

from synapse.media.oembed import OEmbedProvider, OEmbedResult
from synapse.server import HomeServer
from synapse.types import JsonDict
from synapse.util import Clock

from tests.unittest import HomeserverTestCase

try:
    import lxml
except ImportError:
    lxml = None  # type: ignore[assignment]


class OEmbedTests(HomeserverTestCase):
    if not lxml:
        skip = "url preview feature requires lxml"

    def prepare(self, reactor: MemoryReactor, clock: Clock, hs: HomeServer) -> None:
        self.oembed = OEmbedProvider(hs)

    def parse_response(self, response: JsonDict) -> OEmbedResult:
        return self.oembed.parse_oembed_response(
            "https://test", json.dumps(response).encode("utf-8")
        )

    def test_version(self) -> None:
        """Accept versions that are similar to 1.0 as a string or int (or missing)."""
        version: Any
        for version in ("1.0", 1.0, 1):
            result = self.parse_response({"version": version})
            # An empty Open Graph response is an error, ensure the URL is included.
            self.assertIn("og:url", result.open_graph_result)

        # A missing version should be treated as 1.0.
        result = self.parse_response({"type": "link"})
        self.assertIn("og:url", result.open_graph_result)

        # Invalid versions should be rejected.
        for version in ("2.0", "1", 1.1, 0, None, {}, []):
            result = self.parse_response({"version": version, "type": "link"})
            # An empty Open Graph response is an error, ensure the URL is included.
            self.assertEqual({}, result.open_graph_result)

    def test_cache_age(self) -> None:
        """Ensure a cache-age is parsed properly."""
        cache_age: Any
        # Correct-ish cache ages are allowed.
        for cache_age in ("1", 1.0, 1):
            result = self.parse_response({"cache_age": cache_age})
            self.assertEqual(result.cache_age, 1000)

        # Invalid cache ages are ignored.
        for cache_age in ("invalid", {}):
            result = self.parse_response({"cache_age": cache_age})
            self.assertIsNone(result.cache_age)

        # Cache age is optional.
        result = self.parse_response({})
        self.assertIsNone(result.cache_age)

    @parameterized.expand(
        [
            ("title", "title"),
            ("provider_name", "site_name"),
            ("thumbnail_url", "image"),
        ],
        name_func=lambda func, num, p: f"{func.__name__}_{p.args[0]}",
    )
    def test_property(self, oembed_property: str, open_graph_property: str) -> None:
        """Test properties which must be strings."""
        result = self.parse_response({oembed_property: "test"})
        self.assertIn(f"og:{open_graph_property}", result.open_graph_result)
        self.assertEqual(result.open_graph_result[f"og:{open_graph_property}"], "test")

        result = self.parse_response({oembed_property: 1})
        self.assertNotIn(f"og:{open_graph_property}", result.open_graph_result)

    def test_author_name(self) -> None:
        """Test the author_name property."""
        result = self.parse_response({"author_name": "test"})
        self.assertEqual(result.author_name, "test")

        result = self.parse_response({"author_name": 1})
        self.assertIsNone(result.author_name)

    def test_rich(self) -> None:
        """Test a type of rich."""
        result = self.parse_response({"html": "test<img src='foo'>", "type": "rich"})
        self.assertIn("og:description", result.open_graph_result)
        self.assertIn("og:image", result.open_graph_result)
        self.assertEqual(result.open_graph_result["og:description"], "test")
        self.assertEqual(result.open_graph_result["og:image"], "foo")

        result = self.parse_response({"type": "rich"})
        self.assertNotIn("og:description", result.open_graph_result)

        result = self.parse_response({"html": 1, "type": "rich"})
        self.assertNotIn("og:description", result.open_graph_result)

    def test_photo(self) -> None:
        """Test a type of photo."""
        result = self.parse_response({"url": "test", "type": "photo"})
        self.assertIn("og:image", result.open_graph_result)
        self.assertEqual(result.open_graph_result["og:image"], "test")

        result = self.parse_response({"type": "photo"})
        self.assertNotIn("og:image", result.open_graph_result)

        result = self.parse_response({"url": 1, "type": "photo"})
        self.assertNotIn("og:image", result.open_graph_result)

    def test_video(self) -> None:
        """Test a type of video."""
        result = self.parse_response({"html": "test", "type": "video"})
        self.assertIn("og:type", result.open_graph_result)
        self.assertEqual(result.open_graph_result["og:type"], "video.other")
        self.assertIn("og:description", result.open_graph_result)
        self.assertEqual(result.open_graph_result["og:description"], "test")

        result = self.parse_response({"type": "video"})
        self.assertIn("og:type", result.open_graph_result)
        self.assertEqual(result.open_graph_result["og:type"], "video.other")
        self.assertNotIn("og:description", result.open_graph_result)

        result = self.parse_response({"url": 1, "type": "video"})
        self.assertIn("og:type", result.open_graph_result)
        self.assertEqual(result.open_graph_result["og:type"], "video.other")
        self.assertNotIn("og:description", result.open_graph_result)

    def test_link(self) -> None:
        """Test type of link."""
        result = self.parse_response({"type": "link"})
        self.assertIn("og:type", result.open_graph_result)
        self.assertEqual(result.open_graph_result["og:type"], "website")

    def test_title_html_entities(self) -> None:
        """Test HTML entities in title"""
        result = self.parse_response(
            {"title": "Why JSON isn&#8217;t a Good Configuration Language"}
        )
        self.assertEqual(
            result.open_graph_result["og:title"],
            "Why JSON isn’t a Good Configuration Language",
        )



#
# This file is licensed under the Affero General Public License (AGPL) version 3.
#
# Copyright 2019 The Matrix.org Foundation C.I.C.
# Copyright (C) 2023 New Vector, Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# See the GNU Affero General Public License for more details:
# <https://www.gnu.org/licenses/agpl-3.0.html>.
#
# Originally licensed under the Apache License, Version 2.0:
# <http://www.apache.org/licenses/LICENSE-2.0>.
#
# [This file includes modifications made by New Vector Limited]
#
#

from pyperf import perf_counter

from synapse.types import ISynapseReactor
from synapse.util.caches.lrucache import LruCache


async def main(reactor: ISynapseReactor, loops: int) -> float:
    """
    Benchmark `loops` number of insertions into LruCache where half of them are
    evicted.
    """
    cache: LruCache[int, bool] = LruCache(loops // 2)

    start = perf_counter()

    for i in range(loops):
        cache[i] = True

    end = perf_counter() - start

    return end



#
# This file is licensed under the Affero General Public License (AGPL) version 3.
#
# Copyright 2020 The Matrix.org Foundation C.I.C.
# Copyright (C) 2023 New Vector, Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# See the GNU Affero General Public License for more details:
# <https://www.gnu.org/licenses/agpl-3.0.html>.
#
# Originally licensed under the Apache License, Version 2.0:
# <http://www.apache.org/licenses/LICENSE-2.0>.
#
# [This file includes modifications made by New Vector Limited]
#
#
import email.utils
import logging
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Collection,
    Dict,
    Generator,
    Iterable,
    List,
    Mapping,
    Optional,
    Tuple,
    TypeVar,
    Union,
)

import attr
import jinja2
from typing_extensions import Concatenate, ParamSpec

from twisted.internet import defer
from twisted.internet.interfaces import IDelayedCall
from twisted.web.resource import Resource

from synapse.api import errors
from synapse.api.constants import ProfileFields
from synapse.api.errors import SynapseError
from synapse.api.presence import UserPresenceState
from synapse.config import ConfigError
from synapse.events import EventBase
from synapse.events.presence_router import (
    GET_INTERESTED_USERS_CALLBACK,
    GET_USERS_FOR_STATES_CALLBACK,
    PresenceRouter,
)
from synapse.events.utils import ADD_EXTRA_FIELDS_TO_UNSIGNED_CLIENT_EVENT_CALLBACK
from synapse.handlers.account_data import ON_ACCOUNT_DATA_UPDATED_CALLBACK
from synapse.handlers.auth import (
    CHECK_3PID_AUTH_CALLBACK,
    CHECK_AUTH_CALLBACK,
    GET_DISPLAYNAME_FOR_REGISTRATION_CALLBACK,
    GET_USERNAME_FOR_REGISTRATION_CALLBACK,
    IS_3PID_ALLOWED_CALLBACK,
    ON_LOGGED_OUT_CALLBACK,
    AuthHandler,
)
from synapse.handlers.device import DeviceHandler
from synapse.handlers.push_rules import RuleSpec, check_actions
from synapse.http.client import SimpleHttpClient
from synapse.http.server import (
    DirectServeHtmlResource,
    DirectServeJsonResource,
    respond_with_html,
)
from synapse.http.servlet import parse_json_object_from_request
from synapse.http.site import SynapseRequest
from synapse.logging.context import (
    defer_to_thread,
    make_deferred_yieldable,
    run_in_background,
)
from synapse.metrics.background_process_metrics import run_as_background_process
from synapse.module_api.callbacks.account_validity_callbacks import (
    IS_USER_EXPIRED_CALLBACK,
    ON_LEGACY_ADMIN_REQUEST,
    ON_LEGACY_RENEW_CALLBACK,
    ON_LEGACY_SEND_MAIL_CALLBACK,
    ON_USER_LOGIN_CALLBACK,
    ON_USER_REGISTRATION_CALLBACK,
)
from synapse.module_api.callbacks.spamchecker_callbacks import (
    CHECK_EVENT_FOR_SPAM_CALLBACK,
    CHECK_LOGIN_FOR_SPAM_CALLBACK,
    CHECK_MEDIA_FILE_FOR_SPAM_CALLBACK,
    CHECK_REGISTRATION_FOR_SPAM_CALLBACK,
    CHECK_USERNAME_FOR_SPAM_CALLBACK,
    SHOULD_DROP_FEDERATED_EVENT_CALLBACK,
    USER_MAY_CREATE_ROOM_ALIAS_CALLBACK,
    USER_MAY_CREATE_ROOM_CALLBACK,
    USER_MAY_INVITE_CALLBACK,
    USER_MAY_JOIN_ROOM_CALLBACK,
    USER_MAY_PUBLISH_ROOM_CALLBACK,
    USER_MAY_SEND_3PID_INVITE_CALLBACK,
    SpamCheckerModuleApiCallbacks,
)
from synapse.module_api.callbacks.third_party_event_rules_callbacks import (
    CHECK_CAN_DEACTIVATE_USER_CALLBACK,
    CHECK_CAN_SHUTDOWN_ROOM_CALLBACK,
    CHECK_EVENT_ALLOWED_CALLBACK,
    CHECK_THREEPID_CAN_BE_INVITED_CALLBACK,
    CHECK_VISIBILITY_CAN_BE_MODIFIED_CALLBACK,
    ON_ADD_USER_THIRD_PARTY_IDENTIFIER_CALLBACK,
    ON_CREATE_ROOM_CALLBACK,
    ON_NEW_EVENT_CALLBACK,
    ON_PROFILE_UPDATE_CALLBACK,
    ON_REMOVE_USER_THIRD_PARTY_IDENTIFIER_CALLBACK,
    ON_THREEPID_BIND_CALLBACK,
    ON_USER_DEACTIVATION_STATUS_CHANGED_CALLBACK,
)
from synapse.push.httppusher import HttpPusher
from synapse.rest.client.login import LoginResponse
from synapse.storage import DataStore
from synapse.storage.background_updates import (
    DEFAULT_BATCH_SIZE_CALLBACK,
    MIN_BATCH_SIZE_CALLBACK,
    ON_UPDATE_CALLBACK,
)
from synapse.storage.database import DatabasePool, LoggingTransaction
from synapse.storage.databases.main.roommember import ProfileInfo
from synapse.types import (
    DomainSpecificString,
    JsonDict,
    JsonMapping,
    Requester,
    RoomAlias,
    RoomID,
    StateMap,
    UserID,
    UserInfo,
    UserProfile,
    create_requester,
)
from synapse.types.state import StateFilter
from synapse.util import Clock
from synapse.util.async_helpers import maybe_awaitable
from synapse.util.caches.descriptors import CachedFunction, cached as _cached
from synapse.util.frozenutils import freeze

if TYPE_CHECKING:
    from synapse.app.generic_worker import GenericWorkerStore
    from synapse.server import HomeServer


T = TypeVar("T")
P = ParamSpec("P")
F = TypeVar("F", bound=Callable[..., Any])

"""
This package defines the 'stable' API which can be used by extension modules which
are loaded into Synapse.
"""

PRESENCE_ALL_USERS = PresenceRouter.ALL_USERS
NOT_SPAM = SpamCheckerModuleApiCallbacks.NOT_SPAM

__all__ = [
    "errors",
    "make_deferred_yieldable",
    "parse_json_object_from_request",
    "respond_with_html",
    "run_in_background",
    "run_as_background_process",
    "cached",
    "NOT_SPAM",
    "UserID",
    "DatabasePool",
    "LoggingTransaction",
    "DirectServeHtmlResource",
    "DirectServeJsonResource",
    "ModuleApi",
    "PRESENCE_ALL_USERS",
    "LoginResponse",
    "JsonDict",
    "JsonMapping",
    "EventBase",
    "StateMap",
    "ProfileInfo",
    "RoomAlias",
    "UserProfile",
]

logger = logging.getLogger(__name__)


@attr.s(auto_attribs=True)
class UserIpAndAgent:
    """
    An IP address and user agent used by a user to connect to this homeserver.
    """

    ip: str
    user_agent: str
    # The time at which this user agent/ip was last seen.
    last_seen: int


def cached(
    *,
    max_entries: int = 1000,
    num_args: Optional[int] = None,
    uncached_args: Optional[Collection[str]] = None,
) -> Callable[[F], CachedFunction[F]]:
    """Returns a decorator that applies a memoizing cache around the function. This
    decorator behaves similarly to functools.lru_cache.

    Example:

        @cached()
        def foo('a', 'b'):
            ...

    Added in Synapse v1.74.0.

    Args:
        max_entries: The maximum number of entries in the cache. If the cache is full
            and a new entry is added, the least recently accessed entry will be evicted
            from the cache.
        num_args: The number of positional arguments (excluding `self`) to use as cache
            keys. Defaults to all named args of the function.
        uncached_args: A list of argument names to not use as the cache key. (`self` is
            always ignored.) Cannot be used with num_args.

    Returns:
        A decorator that applies a memoizing cache around the function.
    """
    return _cached(
        max_entries=max_entries,
        num_args=num_args,
        uncached_args=uncached_args,
    )


class ModuleApi:
    """A proxy object that gets passed to various plugin modules so they
    can register new users etc if necessary.
    """

    def __init__(self, hs: "HomeServer", auth_handler: AuthHandler) -> None:
        self._hs = hs

        # TODO: Fix this type hint once the types for the data stores have been ironed
        #       out.
        self._store: Union[DataStore, "GenericWorkerStore"] = hs.get_datastores().main
        self._storage_controllers = hs.get_storage_controllers()
        self._auth = hs.get_auth()
        self._auth_handler = auth_handler
        self._server_name = hs.hostname
        self._presence_stream = hs.get_event_sources().sources.presence
        self._state = hs.get_state_handler()
        self._clock: Clock = hs.get_clock()
        self._registration_handler = hs.get_registration_handler()
        self._send_email_handler = hs.get_send_email_handler()
        self._push_rules_handler = hs.get_push_rules_handler()
        self._pusherpool = hs.get_pusherpool()
        self._device_handler = hs.get_device_handler()
        self.custom_template_dir = hs.config.server.custom_template_directory
        self._callbacks = hs.get_module_api_callbacks()
        self.msc3861_oauth_delegation_enabled = hs.config.experimental.msc3861.enabled
        self._event_serializer = hs.get_event_client_serializer()

        try:
            app_name = self._hs.config.email.email_app_name

            self._from_string = self._hs.config.email.email_notif_from % {
                "app": app_name
            }
        except (KeyError, TypeError):
            # If substitution failed (which can happen if the string contains
            # placeholders other than just "app", or if the type of the placeholder is
            # not a string), fall back to the bare strings.
            self._from_string = self._hs.config.email.email_notif_from

        self._raw_from = email.utils.parseaddr(self._from_string)[1]

        # We expose these as properties below in order to attach a helpful docstring.
        self._http_client: SimpleHttpClient = hs.get_simple_http_client()
        self._public_room_list_manager = PublicRoomListManager(hs)
        self._account_data_manager = AccountDataManager(hs)

        self._password_auth_provider = hs.get_password_auth_provider()
        self._presence_router = hs.get_presence_router()
        self._account_data_handler = hs.get_account_data_handler()

    #################################################################################
    # The following methods should only be called during the module's initialisation.

    def register_spam_checker_callbacks(
        self,
        *,
        check_event_for_spam: Optional[CHECK_EVENT_FOR_SPAM_CALLBACK] = None,
        should_drop_federated_event: Optional[
            SHOULD_DROP_FEDERATED_EVENT_CALLBACK
        ] = None,
        user_may_join_room: Optional[USER_MAY_JOIN_ROOM_CALLBACK] = None,
        user_may_invite: Optional[USER_MAY_INVITE_CALLBACK] = None,
        user_may_send_3pid_invite: Optional[USER_MAY_SEND_3PID_INVITE_CALLBACK] = None,
        user_may_create_room: Optional[USER_MAY_CREATE_ROOM_CALLBACK] = None,
        user_may_create_room_alias: Optional[
            USER_MAY_CREATE_ROOM_ALIAS_CALLBACK
        ] = None,
        user_may_publish_room: Optional[USER_MAY_PUBLISH_ROOM_CALLBACK] = None,
        check_username_for_spam: Optional[CHECK_USERNAME_FOR_SPAM_CALLBACK] = None,
        check_registration_for_spam: Optional[
            CHECK_REGISTRATION_FOR_SPAM_CALLBACK
        ] = None,
        check_media_file_for_spam: Optional[CHECK_MEDIA_FILE_FOR_SPAM_CALLBACK] = None,
        check_login_for_spam: Optional[CHECK_LOGIN_FOR_SPAM_CALLBACK] = None,
    ) -> None:
        """Registers callbacks for spam checking capabilities.

        Added in Synapse v1.37.0.
        """
        return self._callbacks.spam_checker.register_callbacks(
            check_event_for_spam=check_event_for_spam,
            should_drop_federated_event=should_drop_federated_event,
            user_may_join_room=user_may_join_room,
            user_may_invite=user_may_invite,
            user_may_send_3pid_invite=user_may_send_3pid_invite,
            user_may_create_room=user_may_create_room,
            user_may_create_room_alias=user_may_create_room_alias,
            user_may_publish_room=user_may_publish_room,
            check_username_for_spam=check_username_for_spam,
            check_registration_for_spam=check_registration_for_spam,
            check_media_file_for_spam=check_media_file_for_spam,
            check_login_for_spam=check_login_for_spam,
        )

    def register_account_validity_callbacks(
        self,
        *,
        is_user_expired: Optional[IS_USER_EXPIRED_CALLBACK] = None,
        on_user_registration: Optional[ON_USER_REGISTRATION_CALLBACK] = None,
        on_user_login: Optional[ON_USER_LOGIN_CALLBACK] = None,
        on_legacy_send_mail: Optional[ON_LEGACY_SEND_MAIL_CALLBACK] = None,
        on_legacy_renew: Optional[ON_LEGACY_RENEW_CALLBACK] = None,
        on_legacy_admin_request: Optional[ON_LEGACY_ADMIN_REQUEST] = None,
    ) -> None:
        """Registers callbacks for account validity capabilities.

        Added in Synapse v1.39.0.
        """
        return self._callbacks.account_validity.register_callbacks(
            is_user_expired=is_user_expired,
            on_user_registration=on_user_registration,
            on_user_login=on_user_login,
            on_legacy_send_mail=on_legacy_send_mail,
            on_legacy_renew=on_legacy_renew,
            on_legacy_admin_request=on_legacy_admin_request,
        )

    def register_third_party_rules_callbacks(
        self,
        *,
        check_event_allowed: Optional[CHECK_EVENT_ALLOWED_CALLBACK] = None,
        on_create_room: Optional[ON_CREATE_ROOM_CALLBACK] = None,
        check_threepid_can_be_invited: Optional[
            CHECK_THREEPID_CAN_BE_INVITED_CALLBACK
        ] = None,
        check_visibility_can_be_modified: Optional[
            CHECK_VISIBILITY_CAN_BE_MODIFIED_CALLBACK
        ] = None,
        on_new_event: Optional[ON_NEW_EVENT_CALLBACK] = None,
        check_can_shutdown_room: Optional[CHECK_CAN_SHUTDOWN_ROOM_CALLBACK] = None,
        check_can_deactivate_user: Optional[CHECK_CAN_DEACTIVATE_USER_CALLBACK] = None,
        on_profile_update: Optional[ON_PROFILE_UPDATE_CALLBACK] = None,
        on_user_deactivation_status_changed: Optional[
            ON_USER_DEACTIVATION_STATUS_CHANGED_CALLBACK
        ] = None,
        on_threepid_bind: Optional[ON_THREEPID_BIND_CALLBACK] = None,
        on_add_user_third_party_identifier: Optional[
            ON_ADD_USER_THIRD_PARTY_IDENTIFIER_CALLBACK
        ] = None,
        on_remove_user_third_party_identifier: Optional[
            ON_REMOVE_USER_THIRD_PARTY_IDENTIFIER_CALLBACK
        ] = None,
    ) -> None:
        """Registers callbacks for third party event rules capabilities.

        Added in Synapse v1.39.0.
        """
        return self._callbacks.third_party_event_rules.register_third_party_rules_callbacks(
            check_event_allowed=check_event_allowed,
            on_create_room=on_create_room,
            check_threepid_can_be_invited=check_threepid_can_be_invited,
            check_visibility_can_be_modified=check_visibility_can_be_modified,
            on_new_event=on_new_event,
            check_can_shutdown_room=check_can_shutdown_room,
            check_can_deactivate_user=check_can_deactivate_user,
            on_profile_update=on_profile_update,
            on_user_deactivation_status_changed=on_user_deactivation_status_changed,
            on_threepid_bind=on_threepid_bind,
            on_add_user_third_party_identifier=on_add_user_third_party_identifier,
            on_remove_user_third_party_identifier=on_remove_user_third_party_identifier,
        )

    def register_presence_router_callbacks(
        self,
        *,
        get_users_for_states: Optional[GET_USERS_FOR_STATES_CALLBACK] = None,
        get_interested_users: Optional[GET_INTERESTED_USERS_CALLBACK] = None,
    ) -> None:
        """Registers callbacks for presence router capabilities.

        Added in Synapse v1.42.0.
        """
        return self._presence_router.register_presence_router_callbacks(
            get_users_for_states=get_users_for_states,
            get_interested_users=get_interested_users,
        )

    def register_password_auth_provider_callbacks(
        self,
        *,
        check_3pid_auth: Optional[CHECK_3PID_AUTH_CALLBACK] = None,
        on_logged_out: Optional[ON_LOGGED_OUT_CALLBACK] = None,
        auth_checkers: Optional[
            Dict[Tuple[str, Tuple[str, ...]], CHECK_AUTH_CALLBACK]
        ] = None,
        is_3pid_allowed: Optional[IS_3PID_ALLOWED_CALLBACK] = None,
        get_username_for_registration: Optional[
            GET_USERNAME_FOR_REGISTRATION_CALLBACK
        ] = None,
        get_displayname_for_registration: Optional[
            GET_DISPLAYNAME_FOR_REGISTRATION_CALLBACK
        ] = None,
    ) -> None:
        """Registers callbacks for password auth provider capabilities.

        Added in Synapse v1.46.0.
        """
        if self.msc3861_oauth_delegation_enabled:
            raise ConfigError(
                "Cannot use password auth provider callbacks when OAuth delegation is enabled"
            )

        return self._password_auth_provider.register_password_auth_provider_callbacks(
            check_3pid_auth=check_3pid_auth,
            on_logged_out=on_logged_out,
            is_3pid_allowed=is_3pid_allowed,
            auth_checkers=auth_checkers,
            get_username_for_registration=get_username_for_registration,
            get_displayname_for_registration=get_displayname_for_registration,
        )

    def register_background_update_controller_callbacks(
        self,
        *,
        on_update: ON_UPDATE_CALLBACK,
        default_batch_size: Optional[DEFAULT_BATCH_SIZE_CALLBACK] = None,
        min_batch_size: Optional[MIN_BATCH_SIZE_CALLBACK] = None,
    ) -> None:
        """Registers background update controller callbacks.

        Added in Synapse v1.49.0.
        """

        for db in self._hs.get_datastores().databases:
            db.updates.register_update_controller_callbacks(
                on_update=on_update,
                default_batch_size=default_batch_size,
                min_batch_size=min_batch_size,
            )

    def register_account_data_callbacks(
        self,
        *,
        on_account_data_updated: Optional[ON_ACCOUNT_DATA_UPDATED_CALLBACK] = None,
    ) -> None:
        """Registers account data callbacks.

        Added in Synapse 1.57.0.
        """
        return self._account_data_handler.register_module_callbacks(
            on_account_data_updated=on_account_data_updated,
        )

    def register_web_resource(self, path: str, resource: Resource) -> None:
        """Registers a web resource to be served at the given path.

        This function should be called during initialisation of the module.

        If multiple modules register a resource for the same path, the module that
        appears the highest in the configuration file takes priority.

        Added in Synapse v1.37.0.

        Args:
            path: The path to register the resource for.
            resource: The resource to attach to this path.
        """
        self._hs.register_module_web_resource(path, resource)

    def register_add_extra_fields_to_unsigned_client_event_callbacks(
        self,
        *,
        add_field_to_unsigned_callback: Optional[
            ADD_EXTRA_FIELDS_TO_UNSIGNED_CLIENT_EVENT_CALLBACK
        ] = None,
    ) -> None:
        """Registers a callback that can be used to add fields to the unsigned
        section of events.

        The callback is called every time an event is sent down to a client.

        Added in Synapse 1.96.0
        """
        if add_field_to_unsigned_callback is not None:
            self._event_serializer.register_add_extra_fields_to_unsigned_client_event_callback(
                add_field_to_unsigned_callback
            )

    #########################################################################
    # The following methods can be called by the module at any point in time.

    @property
    def http_client(self) -> SimpleHttpClient:
        """Allows making outbound HTTP requests to remote resources.

        An instance of synapse.http.client.SimpleHttpClient

        Added in Synapse v1.22.0.
        """
        return self._http_client

    @property
    def public_room_list_manager(self) -> "PublicRoomListManager":
        """Allows adding to, removing from and checking the status of rooms in the
        public room list.

        An instance of synapse.module_api.PublicRoomListManager

        Added in Synapse v1.22.0.
        """
        return self._public_room_list_manager

    @property
    def account_data_manager(self) -> "AccountDataManager":
        """Allows reading and modifying users' account data.

        Added in Synapse v1.57.0.
        """
        return self._account_data_manager

    @property
    def public_baseurl(self) -> str:
        """The configured public base URL for this homeserver.

        Added in Synapse v1.39.0.
        """
        return self._hs.config.server.public_baseurl

    @property
    def email_app_name(self) -> str:
        """The application name configured in the homeserver's configuration.

        Added in Synapse v1.39.0.
        """
        return self._hs.config.email.email_app_name

    @property
    def server_name(self) -> str:
        """The server name for the local homeserver.

        Added in Synapse v1.53.0.
        """
        return self._server_name

    @property
    def worker_name(self) -> Optional[str]:
        """The name of the worker this specific instance is running as per the
        "worker_name" configuration setting, or None if it's the main process.

        Added in Synapse v1.53.0.
        """
        return self._hs.config.worker.worker_name

    @property
    def worker_app(self) -> Optional[str]:
        """The name of the worker app this specific instance is running as per the
        "worker_app" configuration setting, or None if it's the main process.

        Added in Synapse v1.53.0.
        """
        return self._hs.config.worker.worker_app

    async def get_userinfo_by_id(self, user_id: str) -> Optional[UserInfo]:
        """Get user info by user_id

        Added in Synapse v1.41.0.

        Args:
            user_id: Fully qualified user id.
        Returns:
            UserInfo object if a user was found, otherwise None
        """
        return await self._store.get_user_by_id(user_id)

    async def get_user_by_req(
        self,
        req: SynapseRequest,
        allow_guest: bool = False,
        allow_expired: bool = False,
    ) -> Requester:
        """Check the access_token provided for a request

        Added in Synapse v1.39.0.

        Args:
            req: Incoming HTTP request
            allow_guest: True if guest users should be allowed. If this
                is False, and the access token is for a guest user, an
                AuthError will be thrown
            allow_expired: True if expired users should be allowed. If this
                is False, and the access token is for an expired user, an
                AuthError will be thrown

        Returns:
            The requester for this request

        Raises:
            InvalidClientCredentialsError: if no user by that token exists,
                or the token is invalid.
        """
        return await self._auth.get_user_by_req(
            req,
            allow_guest,
            allow_expired=allow_expired,
        )

    async def is_user_admin(self, user_id: str) -> bool:
        """Checks if a user is a server admin.

        Added in Synapse v1.39.0.

        Args:
            user_id: The Matrix ID of the user to check.

        Returns:
            True if the user is a server admin, False otherwise.
        """
        return await self._store.is_server_admin(UserID.from_string(user_id))

    async def set_user_admin(self, user_id: str, admin: bool) -> None:
        """Sets if a user is a server admin.

        Added in Synapse v1.56.0.

        Args:
            user_id: The Matrix ID of the user to set admin status for.
            admin: True iff the user is to be a server admin, false otherwise.
        """
        await self._store.set_server_admin(UserID.from_string(user_id), admin)

    def get_qualified_user_id(self, username: str) -> str:
        """Qualify a user id, if necessary

        Takes a user id provided by the user and adds the @ and :domain to
        qualify it, if necessary

        Added in Synapse v0.25.0.

        Args:
            username: provided user id

        Returns:
            qualified @user:id
        """
        if username.startswith("@"):
            return username
        return UserID(username, self._hs.hostname).to_string()

    async def get_profile_for_user(self, localpart: str) -> ProfileInfo:
        """Look up the profile info for the user with the given localpart.

        Added in Synapse v1.39.0.

        Args:
            localpart: The localpart to look up profile information for.

        Returns:
            The profile information (i.e. display name and avatar URL).
        """
        server_name = self._hs.hostname
        user_id = UserID.from_string(f"@{localpart}:{server_name}")
        return await self._store.get_profileinfo(user_id)

    async def get_threepids_for_user(self, user_id: str) -> List[Dict[str, str]]:
        """Look up the threepids (email addresses and phone numbers) associated with the
        given Matrix user ID.

        Added in Synapse v1.39.0.

        Args:
            user_id: The Matrix user ID to look up threepids for.

        Returns:
            A list of threepids, each threepid being represented by a dictionary
            containing a "medium" key which value is "email" for email addresses and
            "msisdn" for phone numbers, and an "address" key which value is the
            threepid's address.
        """
        return [attr.asdict(t) for t in await self._store.user_get_threepids(user_id)]

    def check_user_exists(self, user_id: str) -> "defer.Deferred[Optional[str]]":
        """Check if user exists.

        Added in Synapse v0.25.0.

        Args:
            user_id: Complete @user:id

        Returns:
            Canonical (case-corrected) user_id, or None
               if the user is not registered.
        """
        return defer.ensureDeferred(self._auth_handler.check_user_exists(user_id))

    @defer.inlineCallbacks
    def register(
        self,
        localpart: str,
        displayname: Optional[str] = None,
        emails: Optional[List[str]] = None,
    ) -> Generator["defer.Deferred[Any]", Any, Tuple[str, str]]:
        """Registers a new user with given localpart and optional displayname, emails.

        Also returns an access token for the new user.

        Deprecated: avoid this, as it generates a new device with no way to
        return that device to the user. Prefer separate calls to register_user and
        register_device.

        Added in Synapse v0.25.0.

        Args:
            localpart: The localpart of the new user.
            displayname: The displayname of the new user.
            emails: Emails to bind to the new user.

        Returns:
            a 2-tuple of (user_id, access_token)
        """
        logger.warning(
            "Using deprecated ModuleApi.register which creates a dummy user device."
        )
        user_id = yield self.register_user(localpart, displayname, emails or [])
        _, access_token, _, _ = yield self.register_device(user_id)
        return user_id, access_token

    def register_user(
        self,
        localpart: str,
        displayname: Optional[str] = None,
        emails: Optional[List[str]] = None,
        admin: bool = False,
    ) -> "defer.Deferred[str]":
        """Registers a new user with given localpart and optional displayname, emails.

        Added in Synapse v1.2.0.
        Changed in Synapse v1.56.0: add 'admin' argument to register the user as admin.

        Args:
            localpart: The localpart of the new user.
            displayname: The displayname of the new user.
            emails: Emails to bind to the new user.
            admin: True if the user should be registered as a server admin.

        Raises:
            SynapseError if there is an error performing the registration. Check the
                'errcode' property for more information on the reason for failure

        Returns:
            user_id
        """
        return defer.ensureDeferred(
            self._hs.get_registration_handler().register_user(
                localpart=localpart,
                default_display_name=displayname,
                bind_emails=emails or [],
                admin=admin,
            )
        )

    def register_device(
        self,
        user_id: str,
        device_id: Optional[str] = None,
        initial_display_name: Optional[str] = None,
    ) -> "defer.Deferred[Tuple[str, str, Optional[int], Optional[str]]]":
        """Register a device for a user and generate an access token.

        Added in Synapse v1.2.0.

        Args:
            user_id: full canonical @user:id
            device_id: The device ID to check, or None to generate
                a new one.
            initial_display_name: An optional display name for the
                device.

        Returns:
            Tuple of device ID, access token, access token expiration time and refresh token
        """
        return defer.ensureDeferred(
            self._hs.get_registration_handler().register_device(
                user_id=user_id,
                device_id=device_id,
                initial_display_name=initial_display_name,
            )
        )

    def record_user_external_id(
        self, auth_provider_id: str, remote_user_id: str, registered_user_id: str
    ) -> defer.Deferred:
        """Record a mapping between an external user id from a single sign-on provider
        and a mxid.

        Added in Synapse v1.9.0.

        Args:
            auth_provider: identifier for the remote auth provider, see `sso` and
                `oidc_providers` in the homeserver configuration.

                Note that no error is raised if the provided value is not in the
                homeserver configuration.
            external_id: id on that system
            user_id: complete mxid that it is mapped to
        """
        return defer.ensureDeferred(
            self._store.record_user_external_id(
                auth_provider_id, remote_user_id, registered_user_id
            )
        )

    async def create_login_token(
        self,
        user_id: str,
        duration_in_ms: int = (2 * 60 * 1000),
        auth_provider_id: Optional[str] = None,
        auth_provider_session_id: Optional[str] = None,
    ) -> str:
        """Create a login token suitable for m.login.token authentication

        Added in Synapse v1.69.0.

        Args:
            user_id: gives the ID of the user that the token is for

            duration_in_ms: the time that the token will be valid for

            auth_provider_id: the ID of the SSO IdP that the user used to authenticate
                to get this token, if any. This is encoded in the token so that
                /login can report stats on number of successful logins by IdP.

            auth_provider_session_id: The session ID got during login from the SSO IdP,
                if any.
        """
        return await self._hs.get_auth_handler().create_login_token_for_user_id(
            user_id,
            duration_in_ms,
            auth_provider_id,
            auth_provider_session_id,
        )

    @defer.inlineCallbacks
    def invalidate_access_token(
        self, access_token: str
    ) -> Generator["defer.Deferred[Any]", Any, None]:
        """Invalidate an access token for a user

        Can only be called from the main process.

        Added in Synapse v0.25.0.

        Args:
            access_token: access token

        Returns:
            twisted.internet.defer.Deferred - resolves once the access token
               has been removed.

        Raises:
            synapse.api.errors.AuthError: the access token is invalid
        """
        assert isinstance(
            self._device_handler, DeviceHandler
        ), "invalidate_access_token can only be called on the main process"

        # see if the access token corresponds to a device
        user_info = yield defer.ensureDeferred(
            self._auth.get_user_by_access_token(access_token)
        )
        device_id = user_info.get("device_id")
        user_id = user_info["user"].to_string()
        if device_id:
            # delete the device, which will also delete its access tokens
            yield defer.ensureDeferred(
                self._device_handler.delete_devices(user_id, [device_id])
            )
        else:
            # no associated device. Just delete the access token.
            yield defer.ensureDeferred(
                self._auth_handler.delete_access_token(access_token)
            )

    def run_db_interaction(
        self,
        desc: str,
        func: Callable[Concatenate[LoggingTransaction, P], T],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> "defer.Deferred[T]":
        """Run a function with a database connection

        Added in Synapse v0.25.0.

        Args:
            desc: description for the transaction, for metrics etc
            func: function to be run. Passed a database cursor object
                as well as *args and **kwargs
            *args: positional args to be passed to func
            **kwargs: named args to be passed to func

        Returns:
            Result of func
        """
        # type-ignore: See https://github.com/python/mypy/issues/8862
        return defer.ensureDeferred(
            self._store.db_pool.runInteraction(desc, func, *args, **kwargs)  # type: ignore[arg-type]
        )

    def register_cached_function(self, cached_func: CachedFunction) -> None:
        """Register a cached function that should be invalidated across workers.
        Invalidation local to a worker can be done directly using `cached_func.invalidate`,
        however invalidation that needs to go to other workers needs to call `invalidate_cache`
        on the module API instead.

        Added in Synapse v1.69.0.

        Args:
            cached_function: The cached function that will be registered to receive invalidation
            locally and from other workers.
        """
        self._store.register_external_cached_function(
            f"{cached_func.__module__}.{cached_func.__name__}", cached_func
        )

    async def invalidate_cache(
        self, cached_func: CachedFunction, keys: Tuple[Any, ...]
    ) -> None:
        """Invalidate a cache entry of a cached function across workers. The cached function
        needs to be registered on all workers first with `register_cached_function`.

        Added in Synapse v1.69.0.

        Args:
            cached_function: The cached function that needs an invalidation
            keys: keys of the entry to invalidate, usually matching the arguments of the
            cached function.
        """
        cached_func.invalidate(keys)
        await self._store.send_invalidation_to_replication(
            f"{cached_func.__module__}.{cached_func.__name__}",
            keys,
        )

    async def complete_sso_login_async(
        self,
        registered_user_id: str,
        request: SynapseRequest,
        client_redirect_url: str,
        new_user: bool = False,
        auth_provider_id: str = "<unknown>",
    ) -> None:
        """Complete a SSO login by redirecting the user to a page to confirm whether they
        want their access token sent to `client_redirect_url`, or redirect them to that
        URL with a token directly if the URL matches with one of the whitelisted clients.

        Added in Synapse v1.13.0.

        Args:
            registered_user_id: The MXID that has been registered as a previous step of
                of this SSO login.
            request: The request to respond to.
            client_redirect_url: The URL to which to offer to redirect the user (or to
                redirect them directly if whitelisted).
            new_user: set to true to use wording for the consent appropriate to a user
                who has just registered.
            auth_provider_id: the ID of the SSO IdP which was used to log in. This
                is used to track counts of sucessful logins by IdP.
        """
        await self._auth_handler.complete_sso_login(
            registered_user_id,
            auth_provider_id,
            request,
            client_redirect_url,
            new_user=new_user,
        )

    @defer.inlineCallbacks
    def get_state_events_in_room(
        self, room_id: str, types: Iterable[Tuple[str, Optional[str]]]
    ) -> Generator[defer.Deferred, Any, Iterable[EventBase]]:
        """Gets current state events for the given room.

        (This is exposed for compatibility with the old SpamCheckerApi. We should
        probably deprecate it and replace it with an async method in a subclass.)

        Added in Synapse v1.22.0.

        Args:
            room_id: The room ID to get state events in.
            types: The event type and state key (using None
                to represent 'any') of the room state to acquire.

        Returns:
            The filtered state events in the room.
        """
        state_ids = yield defer.ensureDeferred(
            self._storage_controllers.state.get_current_state_ids(
                room_id=room_id, state_filter=StateFilter.from_types(types)
            )
        )
        state = yield defer.ensureDeferred(self._store.get_events(state_ids.values()))
        return state.values()

    async def update_room_membership(
        self,
        sender: str,
        target: str,
        room_id: str,
        new_membership: str,
        content: Optional[JsonDict] = None,
        remote_room_hosts: Optional[List[str]] = None,
    ) -> EventBase:
        """Updates the membership of a user to the given value.

        Added in Synapse v1.46.0.
        Changed in Synapse v1.65.0: Added the 'remote_room_hosts' parameter.

        Args:
            sender: The user performing the membership change. Must be a user local to
                this homeserver.
            target: The user whose membership is changing. This is often the same value
                as `sender`, but it might differ in some cases (e.g. when kicking a user,
                the `sender` is the user performing the kick and the `target` is the user
                being kicked).
            room_id: The room in which to change the membership.
            new_membership: The new membership state of `target` after this operation. See
                https://spec.matrix.org/unstable/client-server-api/#mroommember for the
                list of allowed values.
            content: Additional values to include in the resulting event's content.
            remote_room_hosts: Remote servers to use for remote joins/knocks/etc.

        Returns:
            The newly created membership event.

        Raises:
            RuntimeError if the `sender` isn't a local user.
            ShadowBanError if a shadow-banned requester attempts to send an invite.
            SynapseError if the module attempts to send a membership event that isn't
                allowed, either by the server's configuration (e.g. trying to set a
                per-room display name that's too long) or by the validation rules around
                membership updates (e.g. the `membership` value is invalid).
        """
        if not self.is_mine(sender):
            raise RuntimeError(
                "Tried to send an event as a user that isn't local to this homeserver",
            )

        requester = create_requester(sender)
        target_user_id = UserID.from_string(target)

        if content is None:
            content = {}

        # Set the profile if not already done by the module.
        if (
            ProfileFields.AVATAR_URL not in content
            or ProfileFields.DISPLAYNAME not in content
        ):
            try:
                # Try to fetch the user's profile.
                profile = await self._hs.get_profile_handler().get_profile(
                    target_user_id.to_string(),
                )
            except SynapseError as e:
                # If the profile couldn't be found, use default values.
                profile = {
                    ProfileFields.DISPLAYNAME: target_user_id.localpart,
                    ProfileFields.AVATAR_URL: None,
                }

                if e.code != 404:
                    # If the error isn't 404, it means we tried to fetch the profile over
                    # federation but the remote server responded with a non-standard
                    # status code.
                    logger.error(
                        "Got non-404 error status when fetching profile for %s",
                        target_user_id.to_string(),
                    )

            # Set the profile where it needs to be set.
            for field_name in [ProfileFields.AVATAR_URL, ProfileFields.DISPLAYNAME]:
                if field_name not in content and field_name in profile:
                    content[field_name] = profile[field_name]

        event_id, _ = await self._hs.get_room_member_handler().update_membership(
            requester=requester,
            target=target_user_id,
            room_id=room_id,
            action=new_membership,
            content=content,
            remote_room_hosts=remote_room_hosts,
        )

        # Try to retrieve the resulting event.
        event = await self._hs.get_datastores().main.get_event(event_id)

        return event

    async def create_and_send_event_into_room(self, event_dict: JsonDict) -> EventBase:
        """Create and send an event into a room.

        Membership events are not supported by this method. To update a user's membership
        in a room, please use the `update_room_membership` method instead.

        Added in Synapse v1.22.0.

        Args:
            event_dict: A dictionary representing the event to send.
                Required keys are `type`, `room_id`, `sender` and `content`.

        Returns:
            The event that was sent. If state event deduplication happened, then
                the previous, duplicate event instead.

        Raises:
            SynapseError if the event was not allowed.
        """
        # Create a requester object
        requester = create_requester(
            event_dict["sender"], authenticated_entity=self._server_name
        )

        # Create and send the event
        (
            event,
            _,
        ) = await self._hs.get_event_creation_handler().create_and_send_nonmember_event(
            requester,
            event_dict,
            ratelimit=False,
            ignore_shadow_ban=True,
        )

        return event

    async def send_local_online_presence_to(self, users: Iterable[str]) -> None:
        """
        Forces the equivalent of a presence initial_sync for a set of local or remote
        users. The users will receive presence for all currently online users that they
        are considered interested in.

        Updates to remote users will be sent immediately, whereas local users will receive
        them on their next sync attempt.

        Note that this method can only be run on the process that is configured to write to the
        presence stream. By default this is the main process.

        Added in Synapse v1.32.0.
        """
        if self._hs._instance_name not in self._hs.config.worker.writers.presence:
            raise Exception(
                "send_local_online_presence_to can only be run "
                "on the process that is configured to write to the "
                "presence stream (by default this is the main process)",
            )

        local_users = set()
        remote_users = set()
        for user in users:
            if self._hs.is_mine_id(user):
                local_users.add(user)
            else:
                remote_users.add(user)

        # We pull out the presence handler here to break a cyclic
        # dependency between the presence router and module API.
        presence_handler = self._hs.get_presence_handler()

        if local_users:
            # Force a presence initial_sync for these users next time they sync.
            await presence_handler.send_full_presence_to_users(local_users)

        for user in remote_users:
            # Retrieve presence state for currently online users that this user
            # is considered interested in.
            presence_events, _ = await self._presence_stream.get_new_events(
                UserID.from_string(user), from_key=None, include_offline=False
            )

            # Send to remote destinations.
            destination = UserID.from_string(user).domain
            await presence_handler.get_federation_queue().send_presence_to_destinations(
                presence_events, [destination]
            )

    async def set_presence_for_users(
        self, users: Mapping[str, Tuple[str, Optional[str]]]
    ) -> None:
        """
        Update the internal presence state of users.

        This can be used for either local or remote users.

        Note that this method can only be run on the process that is configured to write to the
        presence stream. By default, this is the main process.

        Added in Synapse v1.96.0.
        """

        # We pull out the presence handler here to break a cyclic
        # dependency between the presence router and module API.
        presence_handler = self._hs.get_presence_handler()

        from synapse.handlers.presence import PresenceHandler

        assert isinstance(presence_handler, PresenceHandler)

        states = await presence_handler.current_state_for_users(users.keys())
        for user_id, (state, status_msg) in users.items():
            prev_state = states.setdefault(user_id, UserPresenceState.default(user_id))
            states[user_id] = prev_state.copy_and_replace(
                state=state, status_msg=status_msg
            )

        await presence_handler._update_states(states.values(), force_notify=True)

    def looping_background_call(
        self,
        f: Callable,
        msec: float,
        *args: object,
        desc: Optional[str] = None,
        run_on_all_instances: bool = False,
        **kwargs: object,
    ) -> None:
        """Wraps a function as a background process and calls it repeatedly.

        NOTE: Will only run on the instance that is configured to run
        background processes (which is the main process by default), unless
        `run_on_all_workers` is set.

        Waits `msec` initially before calling `f` for the first time.

        Added in Synapse v1.39.0.

        Args:
            f: The function to call repeatedly. f can be either synchronous or
                asynchronous, and must follow Synapse's logcontext rules.
                More info about logcontexts is available at
                https://element-hq.github.io/synapse/latest/log_contexts.html
            msec: How long to wait between calls in milliseconds.
            *args: Positional arguments to pass to function.
            desc: The background task's description. Default to the function's name.
            run_on_all_instances: Whether to run this on all instances, rather
                than just the instance configured to run background tasks.
            **kwargs: Key arguments to pass to function.
        """
        if desc is None:
            desc = f.__name__

        if self._hs.config.worker.run_background_tasks or run_on_all_instances:
            self._clock.looping_call(
                run_as_background_process,
                msec,
                desc,
                lambda: maybe_awaitable(f(*args, **kwargs)),
            )
        else:
            logger.warning(
                "Not running looping call %s as the configuration forbids it",
                f,
            )

    def should_run_background_tasks(self) -> bool:
        """
        Return true if and only if the current worker is configured to run
        background tasks.
        There should only be one worker configured to run background tasks, so
        this is helpful when you need to only run a task on one worker but don't
        have any other good way to choose which one.

        Added in Synapse v1.89.0.
        """
        return self._hs.config.worker.run_background_tasks

    def delayed_background_call(
        self,
        msec: float,
        f: Callable,
        *args: object,
        desc: Optional[str] = None,
        **kwargs: object,
    ) -> IDelayedCall:
        """Wraps a function as a background process and calls it in a given number of milliseconds.

        The scheduled call is not persistent: if the current Synapse instance is
        restarted before the call is made, the call will not be made.

        Added in Synapse v1.90.0.

        Args:
            msec: How long to wait before calling, in milliseconds.
            f: The function to call once. f can be either synchronous or
                asynchronous, and must follow Synapse's logcontext rules.
                More info about logcontexts is available at
                https://element-hq.github.io/synapse/latest/log_contexts.html
            *args: Positional arguments to pass to function.
            desc: The background task's description. Default to the function's name.
            **kwargs: Keyword arguments to pass to function.

        Returns:
            IDelayedCall handle from twisted, which allows to cancel the delayed call if desired.
        """

        if desc is None:
            desc = f.__name__

        return self._clock.call_later(
            # convert ms to seconds as needed by call_later.
            msec * 0.001,
            run_as_background_process,
            desc,
            lambda: maybe_awaitable(f(*args, **kwargs)),
        )

    async def sleep(self, seconds: float) -> None:
        """Sleeps for the given number of seconds.

        Added in Synapse v1.49.0.
        """

        await self._clock.sleep(seconds)

    async def send_http_push_notification(
        self,
        user_id: str,
        device_id: Optional[str],
        content: JsonDict,
        tweaks: Optional[JsonMapping] = None,
        default_payload: Optional[JsonMapping] = None,
    ) -> Dict[str, bool]:
        """Send an HTTP push notification that is forwarded to the registered push gateway
        for the specified user/device.

        Added in Synapse v1.82.0.

        Args:
            user_id: The user ID to send the push notification to.
            device_id: The device ID of the device where to send the push notification. If `None`,
            the notification will be sent to all registered HTTP pushers of the user.
            content: A dict of values that will be put in the `notification` field of the push
            (cf Push Gateway spec). `devices` field will be overrided if included.
            tweaks: A dict of `tweaks` that will be inserted in the `devices` section, cf spec.
            default_payload: default payload to add in `devices[0].data.default_payload`.
            This will be merged (and override if some matching values already exist there)
            with existing `default_payload`.

        Returns:
            a dict reprensenting the status of the push per device ID
        """
        status = {}
        if user_id in self._pusherpool.pushers:
            for p in self._pusherpool.pushers[user_id].values():
                if isinstance(p, HttpPusher) and (
                    not device_id or p.device_id == device_id
                ):
                    res = await p.dispatch_push(content, tweaks, default_payload)
                    # Check if the push was successful and no pushers were rejected.
                    sent = res is not False and not res

                    # This is mainly to accomodate mypy
                    # device_id should never be empty after the `set_device_id_for_pushers`
                    # background job has been properly run.
                    if p.device_id:
                        status[p.device_id] = sent
        return status

    async def send_mail(
        self,
        recipient: str,
        subject: str,
        html: str,
        text: str,
    ) -> None:
        """Send an email on behalf of the homeserver.

        Added in Synapse v1.39.0.

        Args:
            recipient: The email address for the recipient.
            subject: The email's subject.
            html: The email's HTML content.
            text: The email's text content.
        """
        await self._send_email_handler.send_email(
            email_address=recipient,
            subject=subject,
            app_name=self.email_app_name,
            html=html,
            text=text,
        )

    def read_templates(
        self,
        filenames: List[str],
        custom_template_directory: Optional[str] = None,
    ) -> List[jinja2.Template]:
        """Read and load the content of the template files at the given location.
        By default, Synapse will look for these templates in its configured template
        directory, but another directory to search in can be provided.

        Added in Synapse v1.39.0.

        Args:
            filenames: The name of the template files to look for.
            custom_template_directory: An additional directory to look for the files in.

        Returns:
            A list containing the loaded templates, with the orders matching the one of
            the filenames parameter.
        """
        return self._hs.config.server.read_templates(
            filenames,
            (td for td in (self.custom_template_dir, custom_template_directory) if td),
        )

    def is_mine(self, id: Union[str, DomainSpecificString]) -> bool:
        """
        Checks whether an ID (user id, room, ...) comes from this homeserver.

        Added in Synapse v1.44.0.

        Args:
            id: any Matrix id (e.g. user id, room id, ...), either as a raw id,
                e.g. string "@user:example.com" or as a parsed UserID, RoomID, ...
        Returns:
            True if id comes from this homeserver, False otherwise.
        """
        if isinstance(id, DomainSpecificString):
            return self._hs.is_mine(id)
        else:
            return self._hs.is_mine_id(id)

    async def get_user_ip_and_agents(
        self, user_id: str, since_ts: int = 0
    ) -> List[UserIpAndAgent]:
        """
        Return the list of user IPs and agents for a user.

        Added in Synapse v1.44.0.

        Args:
            user_id: the id of a user, local or remote
            since_ts: a timestamp in seconds since the epoch,
                or the epoch itself if not specified.
        Returns:
            The list of all UserIpAndAgent that the user has
            used to connect to this homeserver since `since_ts`.
            If the user is remote, this list is empty.
        """
        # Don't hit the db if this is not a local user.
        is_mine = False
        try:
            # Let's be defensive against ill-formed strings.
            if self.is_mine(user_id):
                is_mine = True
        except Exception:
            pass

        if is_mine:
            raw_data = await self._store.get_user_ip_and_agents(
                UserID.from_string(user_id), since_ts
            )
            # Sanitize some of the data. We don't want to return tokens.
            return [
                UserIpAndAgent(
                    ip=data["ip"],
                    user_agent=data["user_agent"],
                    last_seen=data["last_seen"],
                )
                for data in raw_data
            ]
        else:
            return []

    async def get_room_state(
        self,
        room_id: str,
        event_filter: Optional[Iterable[Tuple[str, Optional[str]]]] = None,
    ) -> StateMap[EventBase]:
        """Returns the current state of the given room.

        The events are returned as a mapping, in which the key for each event is a tuple
        which first element is the event's type and the second one is its state key.

        Added in Synapse v1.47.0

        Args:
            room_id: The ID of the room to get state from.
            event_filter: A filter to apply when retrieving events. None if no filter
                should be applied. If provided, must be an iterable of tuples. A tuple's
                first element is the event type and the second is the state key, or is
                None if the state key should not be filtered on.
                An example of a filter is:
                    [
                        ("m.room.member", "@alice:example.com"),  # Member event for @alice:example.com
                        ("org.matrix.some_event", ""),  # State event of type "org.matrix.some_event"
                                                        # with an empty string as its state key
                        ("org.matrix.some_other_event", None),  # State events of type "org.matrix.some_other_event"
                                                                # regardless of their state key
                    ]
        """
        state_filter = None
        if event_filter:
            # If a filter was provided, turn it into a StateFilter and retrieve a filtered
            # view of the state.
            state_filter = StateFilter.from_types(event_filter)

        state_ids = await self._storage_controllers.state.get_current_state_ids(
            room_id,
            state_filter,
        )

        state_events = await self._store.get_events(state_ids.values())

        return {key: state_events[event_id] for key, event_id in state_ids.items()}

    async def defer_to_thread(
        self,
        f: Callable[P, T],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> T:
        """Runs the given function in a separate thread from Synapse's thread pool.

        Added in Synapse v1.49.0.

        Args:
            f: The function to run.
            args: The function's arguments.
            kwargs: The function's keyword arguments.

        Returns:
            The return value of the function once ran in a thread.
        """
        return await defer_to_thread(self._hs.get_reactor(), f, *args, **kwargs)

    async def check_username(self, username: str) -> None:
        """Checks if the provided username uses the grammar defined in the Matrix
        specification, and is already being used by an existing user.

        Added in Synapse v1.52.0.

        Args:
            username: The username to check. This is the local part of the user's full
                Matrix user ID, i.e. it's "alice" if the full user ID is "@alice:foo.com".

        Raises:
            SynapseError with the errcode "M_USER_IN_USE" if the username is already in
            use.
        """
        await self._registration_handler.check_username(username)

    async def store_remote_3pid_association(
        self, user_id: str, medium: str, address: str, id_server: str
    ) -> None:
        """Stores an existing association between a user ID and a third-party identifier.

        The association must already exist on the remote identity server.

        Added in Synapse v1.56.0.

        Args:
            user_id: The user ID that's been associated with the 3PID.
            medium: The medium of the 3PID (current supported values are "msisdn" and
                "email").
            address: The address of the 3PID.
            id_server: The identity server the 3PID association has been registered on.
                This should only be the domain (or IP address, optionally with the port
                number) for the identity server. This will be used to reach out to the
                identity server using HTTPS (unless specified otherwise by Synapse's
                configuration) when attempting to unbind the third-party identifier.


        """
        await self._store.add_user_bound_threepid(user_id, medium, address, id_server)

    def check_push_rule_actions(
        self, actions: List[Union[str, Dict[str, str]]]
    ) -> None:
        """Checks if the given push rule actions are valid according to the Matrix
        specification.

        See https://spec.matrix.org/v1.2/client-server-api/#actions for the list of valid
        actions.

        Added in Synapse v1.58.0.

        Args:
            actions: the actions to check.

        Raises:
            synapse.module_api.errors.InvalidRuleException if the actions are invalid.
        """
        check_actions(actions)

    async def set_push_rule_action(
        self,
        user_id: str,
        scope: str,
        kind: str,
        rule_id: str,
        actions: List[Union[str, Dict[str, str]]],
    ) -> None:
        """Changes the actions of an existing push rule for the given user.

        See https://spec.matrix.org/v1.2/client-server-api/#push-rules for more
        information about push rules and their syntax.

        Can only be called on the main process.

        Added in Synapse v1.58.0.

        Args:
            user_id: the user for which to change the push rule's actions.
            scope: the push rule's scope, currently only "global" is allowed.
            kind: the push rule's kind.
            rule_id: the push rule's identifier.
            actions: the actions to run when the rule's conditions match.

        Raises:
            RuntimeError if this method is called on a worker or `scope` is invalid.
            synapse.module_api.errors.RuleNotFoundException if the rule being modified
                can't be found.
            synapse.module_api.errors.InvalidRuleException if the actions are invalid.
        """
        if self.worker_app is not None:
            raise RuntimeError("module tried to change push rule actions on a worker")

        if scope != "global":
            raise RuntimeError(
                "invalid scope %s, only 'global' is currently allowed" % scope
            )

        spec = RuleSpec(scope, kind, rule_id, "actions")
        await self._push_rules_handler.set_rule_attr(
            user_id, spec, {"actions": actions}
        )

    async def get_monthly_active_users_by_service(
        self, start_timestamp: Optional[int] = None, end_timestamp: Optional[int] = None
    ) -> List[Tuple[str, str]]:
        """Generates list of monthly active users and their services.
        Please see corresponding storage docstring for more details.

        Added in Synapse v1.61.0.

        Arguments:
            start_timestamp: If specified, only include users that were first active
                at or after this point
            end_timestamp: If specified, only include users that were first active
                at or before this point

        Returns:
            A list of tuples (appservice_id, user_id)

        """
        return await self._store.get_monthly_active_users_by_service(
            start_timestamp, end_timestamp
        )

    async def get_canonical_room_alias(self, room_id: RoomID) -> Optional[RoomAlias]:
        """
        Retrieve the given room's current canonical alias.

        A room may declare an alias as "canonical", meaning that it is the
        preferred alias to use when referring to the room. This function
        retrieves that alias from the room's state.

        Added in Synapse v1.86.0.

        Args:
            room_id: The Room ID to find the alias of.

        Returns:
            None if the room ID does not exist, or if the room exists but has no canonical alias.
            Otherwise, the parsed room alias.
        """
        room_alias_str = (
            await self._storage_controllers.state.get_canonical_alias_for_room(
                room_id.to_string()
            )
        )
        if room_alias_str:
            return RoomAlias.from_string(room_alias_str)
        return None

    async def lookup_room_alias(self, room_alias: str) -> Tuple[str, List[str]]:
        """
        Get the room ID associated with a room alias.

        Added in Synapse v1.65.0.

        Args:
            room_alias: The alias to look up.

        Returns:
            A tuple of:
                The room ID (str).
                Hosts likely to be participating in the room ([str]).

        Raises:
            SynapseError if room alias is invalid or could not be found.
        """
        alias = RoomAlias.from_string(room_alias)
        (room_id, hosts) = await self._hs.get_room_member_handler().lookup_room_alias(
            alias
        )

        return room_id.to_string(), hosts

    async def create_room(
        self,
        user_id: str,
        config: JsonDict,
        ratelimit: bool = True,
        creator_join_profile: Optional[JsonDict] = None,
    ) -> Tuple[str, Optional[str]]:
        """Creates a new room.

        Added in Synapse v1.65.0.

        Args:
            user_id:
                The user who requested the room creation.
            config : A dict of configuration options. See "Request body" of:
                https://spec.matrix.org/latest/client-server-api/#post_matrixclientv3createroom
            ratelimit: set to False to disable the rate limiter for this specific operation.

            creator_join_profile:
                Set to override the displayname and avatar for the creating
                user in this room. If unset, displayname and avatar will be
                derived from the user's profile. If set, should contain the
                values to go in the body of the 'join' event (typically
                `avatar_url` and/or `displayname`.

        Returns:
                A tuple containing: 1) the room ID (str), 2) if an alias was requested,
                the room alias (str), otherwise None if no alias was requested.

        Raises:
            ResourceLimitError if server is blocked to some resource being
            exceeded.
            RuntimeError if the user_id does not refer to a local user.
            SynapseError if the user_id is invalid, room ID couldn't be stored, or
            something went horribly wrong.
        """
        if not self.is_mine(user_id):
            raise RuntimeError(
                "Tried to create a room as a user that isn't local to this homeserver",
            )

        requester = create_requester(user_id)
        room_id, room_alias, _ = await self._hs.get_room_creation_handler().create_room(
            requester=requester,
            config=config,
            ratelimit=ratelimit,
            creator_join_profile=creator_join_profile,
        )
        room_alias_str = room_alias.to_string() if room_alias else None
        return room_id, room_alias_str

    async def delete_room(self, room_id: str) -> None:
        """
        Schedules the deletion of a room from Synapse's database.

        If the room is already being deleted, this method does nothing.
        This method does not wait for the room to be deleted.

        Added in Synapse v1.89.0.
        """
        # Future extensions to this method might want to e.g. allow use of `force_purge`.
        # TODO In the future we should make sure this is persistent.
        await self._hs.get_pagination_handler().start_shutdown_and_purge_room(
            room_id,
            {
                "new_room_user_id": None,
                "new_room_name": None,
                "message": None,
                "requester_user_id": None,
                "block": False,
                "purge": True,
                "force_purge": False,
            },
        )

    async def set_displayname(
        self,
        user_id: UserID,
        new_displayname: str,
        deactivation: bool = False,
    ) -> None:
        """Sets a user's display name.

        Added in Synapse v1.76.0.

        Args:
            user_id:
                The user whose display name is to be changed.
            new_displayname:
                The new display name to give the user.
            deactivation:
                Whether this change was made while deactivating the user.
        """
        requester = create_requester(user_id)
        await self._hs.get_profile_handler().set_displayname(
            target_user=user_id,
            requester=requester,
            new_displayname=new_displayname,
            by_admin=True,
            deactivation=deactivation,
        )

    def get_current_time_msec(self) -> int:
        """Returns the current server time in milliseconds."""
        return self._clock.time_msec()


class PublicRoomListManager:
    """Contains methods for adding to, removing from and querying whether a room
    is in the public room list.
    """

    def __init__(self, hs: "HomeServer"):
        self._store = hs.get_datastores().main

    async def room_is_in_public_room_list(self, room_id: str) -> bool:
        """Checks whether a room is in the public room list.

        Added in Synapse v1.22.0.

        Args:
            room_id: The ID of the room.

        Returns:
            Whether the room is in the public room list. Returns False if the room does
            not exist.
        """
        room = await self._store.get_room(room_id)
        if not room:
            return False

        # The first item is whether the room is public.
        return room[0]

    async def add_room_to_public_room_list(self, room_id: str) -> None:
        """Publishes a room to the public room list.

        Added in Synapse v1.22.0.

        Args:
            room_id: The ID of the room.
        """
        await self._store.set_room_is_public(room_id, True)

    async def remove_room_from_public_room_list(self, room_id: str) -> None:
        """Removes a room from the public room list.

        Added in Synapse v1.22.0.

        Args:
            room_id: The ID of the room.
        """
        await self._store.set_room_is_public(room_id, False)


class AccountDataManager:
    """
    Allows modules to manage account data.
    """

    def __init__(self, hs: "HomeServer") -> None:
        self._hs = hs
        self._store = hs.get_datastores().main
        self._handler = hs.get_account_data_handler()

    def _validate_user_id(self, user_id: str) -> None:
        """
        Validates a user ID is valid and local.
        Private method to be used in other account data methods.
        """
        user = UserID.from_string(user_id)
        if not self._hs.is_mine(user):
            raise ValueError(
                f"{user_id} is not local to this homeserver; can't access account data for remote users."
            )

    async def get_global(self, user_id: str, data_type: str) -> Optional[JsonMapping]:
        """
        Gets some global account data, of a specified type, for the specified user.

        The provided user ID must be a valid user ID of a local user.

        Added in Synapse v1.57.0.
        """
        self._validate_user_id(user_id)

        data = await self._store.get_global_account_data_by_type_for_user(
            user_id, data_type
        )
        # We clone and freeze to prevent the module accidentally mutating the
        # dict that lives in the cache, as that could introduce nasty bugs.
        return freeze(data)

    async def put_global(
        self, user_id: str, data_type: str, new_data: JsonDict
    ) -> None:
        """
        Puts some global account data, of a specified type, for the specified user.

        The provided user ID must be a valid user ID of a local user.

        Please note that this will overwrite existing the account data of that type
        for that user!

        Added in Synapse v1.57.0.
        """
        self._validate_user_id(user_id)

        if not isinstance(data_type, str):
            raise TypeError(f"data_type must be a str; got {type(data_type).__name__}")

        if not isinstance(new_data, dict):
            raise TypeError(f"new_data must be a dict; got {type(new_data).__name__}")

        # Ensure the user exists, so we don't just write to users that aren't there.
        if await self._store.get_user_by_id(user_id) is None:
            raise ValueError(f"User {user_id} does not exist on this server.")

        await self._handler.add_account_data_for_user(user_id, data_type, new_data)
